# @Created Date: 2020-09-17 12:02:31 am
# @Filename: database.py
# @Email:  1730416009@stu.suda.edu.cn
# @Author: ZeFeng Zhu
# @Last Modified: 2020-09-27 03:18:19 pm
# @Copyright (c) 2020 MinghuiGroup, Soochow University
from asyncio import sleep as asyncio_sleep
import databases
import orm
import sqlalchemy
from sqlite3 import OperationalError, InterfaceError
from unsync import unsync
from typing import Dict, Iterable
from pdb_profiling.log import Abclog
from random import uniform
# from queue import Queue


class SqliteDB(Abclog):

    def init_table_model(self):
        pass

    def __init__(self, url: str, drop_all: bool = False):
        self.metadata = sqlalchemy.MetaData()
        self.database = databases.Database(url)
        self.engine = sqlalchemy.create_engine(url)
        self.engine.execute("PRAGMA journal_mode=WAL")
        self.init_table_model()
        if drop_all:
            self.metadata.drop_all(self.engine, checkfirst=True)
        self.metadata.create_all(self.engine, checkfirst=True)
        # self.queue = Queue()

    def close(self):
        """TODO: make it real"""
        self.engine.dispose()

    def sync_insert(self, table, values: Iterable[Dict], prefix_with: str = "OR IGNORE"):
        try:
            self.engine.execute(
                table.__table__.insert().prefix_with(prefix_with),
                values)
        except InterfaceError as e:
            self.logger.error(f"{e}\n{values}")
            raise e

    '''
    @unsync
    async def async_insert_queue(self, table, values: Iterable[Dict], prefix_with: str = "OR IGNORE", insert_sleep: float = 30.5):
        self.queue.put_nowait(self.database.execute(table.__table__.insert().values(values).prefix_with(prefix_with)))
        now_task = self.queue.get_nowait()
        await now_task
        self.queue.task_done()
    '''

    @unsync
    async def async_insert(self, table, values: Iterable[Dict], prefix_with: str = "OR IGNORE", insert_sleep_range=(10,30)):
        while True:
            try:
                query = table.__table__.insert().values(values).prefix_with(prefix_with)
                await self.database.execute(query)
                break
            except OperationalError as e:
                insert_sleep = uniform(*insert_sleep_range)
                self.logger.error(f"OperationalError: sleep {insert_sleep}s and try again")
                await asyncio_sleep(insert_sleep)
            except InterfaceError as e:
                self.logger.error(f"{e}\n{values}")
                raise e



# @Created Date: 2020-11-23 10:29:05 am
# @Filename: __init__.py
# @Email:  1730416009@stu.suda.edu.cn
# @Author: ZeFeng Zhu
# @Last Modified: 2020-11-23 10:29:40 am
# @Copyright (c) 2020 MinghuiGroup, Soochow University
import orm
from pdb_profiling.processors.database import SqliteDB


class CustomDB(SqliteDB):

    def init_table_model(self):
        class Mutation(orm.Model):
            __tablename__ = 'Mutation'
            __metadata__ = self.metadata
            __database__ = self.database
            ftId = orm.String(max_length=20, primary_key=True)
            Ref = orm.String(max_length=3, primary_key=True)
            Pos = orm.Integer(primary_key=True)
            Alt = orm.String(max_length=3, primary_key=True)

        class IDMapping(orm.Model):
            __tablename__ = 'IDMapping'
            __metadata__ = self.metadata
            __database__ = self.database
            ftId = orm.String(max_length=20, primary_key=True)
            Entry = orm.String(max_length=50, primary_key=True)
            isoform = orm.String(max_length=50, primary_key=True)
            is_canonical = orm.Boolean()


        self.Mutation = Mutation
        self.IDMapping = IDMapping



# @Created Date: 2019-12-08 06:46:49 pm
# @Filename: process.py
# @Email:  1730416009@stu.suda.edu.cn
# @Author: ZeFeng Zhu
# @Last Modified: 2020-02-16 10:54:32 am
# @Copyright (c) 2020 MinghuiGroup, Soochow University
import re
from pathlib import Path
import pandas as pd
from numpy import nan


class ExtractIsoAlt:
    # pattern_all = re.compile(r"([A-z0-9]+):VAR_SEQ\s([0-9\s]+)\s([A-z\s\->]+)\s\(in\s([^\)]+)\)[^=]+FTId=([A-z0-9_,\s]+)")
    pattern_sep = re.compile(r"([A-z0-9]+):VAR_SEQ\s([0-9\.]+);\s+")
    pattern_iso_Value = re.compile(r'/([^=]+)="([^;]+)"')
    pattern_inIso = re.compile(r"([A-z\s\->]+)\s\(in\s([^\)]+)\)")
    pattern_iso_keyValue = re.compile(r"([^=]+)=([^;]+);\s+")
    pattern_inDe = re.compile(r"([A-z]+)\s->\s([A-z]+)")
    usecols = ["Entry", "Alternative sequence",
               "Alternative products (isoforms)"]

    def __init__(self, path: str, usecols=usecols, sep: str = '\t'):
        pathOb = Path(path)
        prefix = pathOb.stem
        self.dfrm = pd.read_csv(path, sep=sep, usecols=usecols)
        self.dfrm.dropna(inplace=True)
        self.dfrm.drop_duplicates(inplace=True)
        self.dfrm.reset_index(drop=True, inplace=True)
        self.altSeqPath = str(Path(pathOb.parent, f"{prefix}_UNP_AltSeq.tsv"))
        self.altProPath = str(Path(pathOb.parent, f"{prefix}_UNP_AltPro.tsv"))

    def __len__(self):
        return len(self.dfrm)

    def extractAltSeq(self):
        target_col_1 = "Alternative sequence"
        self.dfrm[target_col_1] = self.dfrm.apply(lambda x: x[target_col_1].replace(
            "VAR_SEQ", "{}:VAR_SEQ".format(x["Entry"])), axis=1)
        target_col_2 = "Alternative products (isoforms)"
        self.dfrm[target_col_2] = self.dfrm.apply(
            lambda x: "Entry={}; {}".format(x["Entry"], x[target_col_2]), axis=1)

        altSeq_li = []
        for content in self.dfrm[target_col_1].dropna():
            result = self.pattern_sep.split(content)
            for i in range(1, len(result)-1, 3):
                altSeq_li.append(
                    result[i+2] + '; /Entry="%s"; /AltRange="%s"; ' % tuple(result[i:i+2]))

        altSeq_df = pd.DataFrame(dict(i.groups() for i in self.pattern_iso_Value.finditer(
            content)) for content in altSeq_li)
        altSeq_df.rename(columns={"id": "AltID"}, inplace=True)
        altSeq_df[["AltInfo", "AltIso"]] = altSeq_df.apply(
            lambda x: self.pattern_inIso.search(x["note"]).groups(), result_type='expand', axis=1)
        altSeq_df["AltRange"] = altSeq_df["AltRange"].apply(
            lambda x: [int(i) for i in x.split("..")])

        altSeq_df["AltLen"] = altSeq_df.apply(lambda x: [len(i) for i in self.pattern_inDe.search(
            x["AltInfo"]).groups()] if x["AltInfo"] != "Missing" else len(range(*x["AltRange"]))+1, axis=1)
        altSeq_df.to_csv(self.altSeqPath, sep="\t", index=False)
        self.altSeq_dict = altSeq_df[[
            "AltID", "AltLen", "AltRange"]].to_dict("list")

    @staticmethod
    def getAltProInfo(inputLyst, groupCol="isoforms", flagCol="Name"):
        def toFrame(dict):
            df = pd.DataFrame(dict[groupCol])
            for key, value in dict.items():
                if key == groupCol:
                    continue
                else:
                    df[key] = value
            return df

        outputDict = {groupCol: []}
        flag = 0
        for key, value in inputLyst:
            if not flag and key != flagCol:
                outputDict[key] = value
            if key == flagCol:
                outputDict[groupCol].append({key: value})
                flag += 1
            else:
                if flag:
                    outputDict[groupCol][flag-1][key] = value

        return toFrame(outputDict)

    @staticmethod
    def getAltInterval(alt_id, altSeq_dict):
        if alt_id in ["Displayed", "External", "Not described"]:
            return nan, nan
        elif not alt_id.startswith("VSP"):
            raise ValueError("Unexpected alt_id: %s" % alt_id)
        else:
            mis_info = []
            inDe_info = []
            alt_id_lyst = alt_id.split(", ")
            for altID in alt_id_lyst:
                index = altSeq_dict["AltID"].index(altID)
                len_info = altSeq_dict["AltLen"][index]
                range_info = altSeq_dict["AltRange"][index]
                if isinstance(len_info, int):
                    mis_info.append((range_info, len_info))
                else:
                    inDe_info.append((range_info, len_info))
                    if len_info[0] > len_info[1]:
                        mis_info.append(
                            ([range_info[0], range_info[1]-1], len_info[0]-len_info[1]))

        return mis_info, inDe_info

    @staticmethod
    def getAffectedInterval(mis_info, inDe_info):
        if isinstance(mis_info, float):
            return nan
        affect_Interval = []
        for inDeRange, inDeLen in inDe_info:
            start = inDeRange[0]
            affect_Interval.append((start, start+inDeLen[1]-1))

        for index in range(len(affect_Interval)):
            raw = affect_Interval[index]
            interval = list(raw)
            for misRange, misLen in mis_info:
                if misRange[-1] < raw[0]:
                    interval = [x-misLen for x in interval]
            affect_Interval[index] = interval

        if affect_Interval:
            return affect_Interval
        else:
            return nan

    def extractAltPro(self):
        altPro_df = pd.concat((self.getAltProInfo(self.pattern_iso_keyValue.findall(
            i)) for i in self.dfrm["Alternative products (isoforms)"].dropna()), sort=False)
        altPro_df["AltInterval"] = altPro_df["Sequence"].apply(
            lambda x: self.getAffectedInterval(*self.getAltInterval(x, self.altSeq_dict)))
        altPro_df["UniProt"] = altPro_df.apply(lambda x: x["IsoId"].split(
            ",")[0] if x["Sequence"] != "Displayed" else x["Entry"], axis=1)
        altPro_df.to_csv(self.altProPath, sep="\t", index=False)

    @classmethod
    def main(cls, **kwargs):
        demo = cls(**kwargs)
        if len(demo.dfrm) > 0:
            demo.extractAltSeq()
            demo.extractAltPro()
            return demo.altSeqPath, demo.altProPath
        else:
            return None, None



# @Created Date: 2020-11-13 03:02:36 pm
# @Filename: record.py
# @Email:  1730416009@stu.suda.edu.cn
# @Author: ZeFeng Zhu
# @Last Modified: 2020-11-13 03:02:39 pm
# @Copyright (c) 2020 MinghuiGroup, Soochow University
from unsync import unsync, Unfuture
from typing import Union
from pathlib import Path
from pandas import DataFrame, concat, isna, Series
from numpy import nan
from re import compile as re_compile
from cachetools import LRUCache
from pdb_profiling.utils import init_folder_from_suffix, a_concat, a_read_csv
from pdb_profiling.processors.uniprot import UniProtDB
from pdb_profiling.processors.uniprot.api import UniProtAPI

class UniProts(object):
    
    pattern_sep = re_compile(r"([A-z0-9]+):VAR_SEQ\s([0-9\.]+);\s+")
    pattern_iso_Value = re_compile(r'/([^=]+)="([^;]+)"')
    pattern_inIso = re_compile(r"([A-z\s\->]+)\s\(in\s([^\)]+)\)")
    pattern_iso_keyValue = re_compile(r"([^=]+)=([^;]+);\s+")
    pattern_inDel = re_compile(r"([A-z]+)\s->\s([A-z]+)")
    
    tasks = LRUCache(maxsize=1024)

    @classmethod
    def register_task(cls, key, task):
        cls.tasks[key] = task

    @classmethod
    def set_folder(cls, folder: Union[Path, str]):
        cls.sqlite_api = UniProtDB("sqlite:///%s" % (init_folder_from_suffix(folder, 'local_db')/"uniprot.db"))
    
    @staticmethod
    def getAltProInfo(inputLyst, groupCol="isoforms", flagCol="Name"):
        def toFrame(dict):
            df = DataFrame(dict[groupCol])
            for key, value in dict.items():
                if key == groupCol:
                    continue
                else:
                    df[key] = value
            return df

        outputDict = {groupCol: []}
        flag = 0
        for key, value in inputLyst:
            if not flag and key != flagCol:
                outputDict[key] = value
            if key == flagCol:
                outputDict[groupCol].append({key: value})
                flag += 1
            else:
                if flag:
                    outputDict[groupCol][flag-1][key] = value

        return toFrame(outputDict)

    @staticmethod
    def get_affected_interval(info, ac_len, ac_seq):
        if isinstance(info, float):
            return nan, nan, nan
        n_data = []
        delta_len = 0
        iso_seq = ''
        start_site = 0
        for index in range(len(info)):
            (cur_site, cur_old_end), (cur_old_len, cur_len), after_seq = info[index]
            delta_len += cur_len - cur_old_len
            iso_seq += ac_seq[start_site:cur_site-1] + after_seq
            start_site = cur_old_end
            if index == 0:
                n_data.append((cur_site, cur_site+cur_len-1 if cur_len > 0 else None))
            else:
                (b_site, b_old_end), (b_old_len, _), _ = info[index-1]
                b_start, b_end = n_data[index-1]
                cur_start = (b_end + cur_site - b_old_end) if b_end is not None else (b_start - b_site + cur_site - b_old_len)
                n_data.append((cur_start, cur_start+cur_len-1 if cur_len > 0 else None))
        n_data = [i for i in n_data if i[1] is not None]
        iso_seq += ac_seq[start_site:]
        return (n_data if n_data else nan), ac_len + delta_len, iso_seq

    '''
    @classmethod
    def get_alt_interval(cls, alt_id, alt_seq_dict):
        if alt_id in ("Displayed", "External", "Not described"):
            return nan, nan
        elif not alt_id.startswith("VSP"):
            raise ValueError("Unexpected alt_id: %s" % alt_id)
        else:
            alt_id_lyst = alt_id.split(", ")
            return cls.get_alt_interval_base(alt_id_lyst, alt_seq_dict)
    '''

    @staticmethod
    def get_alt_interval_base(alt_id_lyst, alt_seq_dict):
        info = []
        for altID in alt_id_lyst:
            index = alt_seq_dict["ftId"].index(altID)
            info.append((
                (alt_seq_dict["begin"][index], alt_seq_dict["end"][index]), 
                (alt_seq_dict["before_len"][index], alt_seq_dict["after_len"][index]),
                alt_seq_dict["after"][index]))
        return sorted(info, key=lambda x: x[0][0])
    '''
    @classmethod
    def get_iso_len(cls, ac_len, sequenceStatus, info):
        if sequenceStatus == 'displayed':
            return ac_len
        elif sequenceStatus == 'described':
            _, alt_lens, _ = zip(*info)
            return ac_len + sum(after-before for before, after in alt_lens)
        else:
            return nan
    '''

    '''
    @classmethod
    def deal_with_alternative_pro(cls, dfrm, alt_seq_dict, ac_len, ac_seq):
        assert frozenset(dfrm.columns) >= frozenset(("Entry", "Alternative products (isoforms)", "Sequence", "Length"))
        altPro_df = concat((cls.getAltProInfo(cls.pattern_iso_keyValue.findall(i)) for i in dfrm["Alternative products (isoforms)"].dropna()), sort=False)
        altPro_df["AltInterval"] = altPro_df["Sequence"].apply(lambda x: cls.get_affected_interval(cls.get_alt_interval(x, alt_seq_dict)))
        return altPro_df
    '''

    @staticmethod
    def split_dot_range(x):
        res = x.split("..")
        if len(res) == 2:
            return [int(i) for i in res]
        elif len(res) == 1:
            val = int(res[0])
            return [val, val]
        else:
            raise AssertionError(f"Unexpected range: {x}")

    @classmethod
    @unsync
    def deal_with_alternative_seq(cls, dfrm):
        if isinstance(dfrm, Unfuture):
            dfrm = dfrm.result()
        assert frozenset(dfrm.columns) >= frozenset(("Entry", "Alternative sequence")), str(dfrm.columns)
        if all(dfrm['Alternative sequence'].isnull()):
            return
        dfrm['Alternative sequence'] = dfrm.apply(lambda x: x['Alternative sequence'].replace("VAR_SEQ", f"{x['Entry']}:VAR_SEQ") if not isna(x['Alternative sequence']) else nan, axis=1)
        altSeq_li = []
        for content in dfrm['Alternative sequence'].dropna().to_numpy():
            result = cls.pattern_sep.split(content)
            for i in range(1, len(result)-1, 3):
                altSeq_li.append(result[i+2] + '; /Entry="%s"; /AltRange="%s"; ' % tuple(result[i:i+2]))
        altSeq_df = DataFrame(dict(i.groups() for i in cls.pattern_iso_Value.finditer(content)) for content in altSeq_li)
        altSeq_df.rename(columns={"id": "ftId"}, inplace=True)
        altSeq_df[["AltInfo", "description"]] = altSeq_df.note.apply(lambda x: cls.pattern_inIso.search(x).groups()).apply(Series) 
        altSeq_df[['begin', 'end']] = altSeq_df.AltRange.apply(cls.split_dot_range).apply(Series)
        altSeq_df[["before", "after"]] = altSeq_df.AltInfo.apply(lambda x: cls.pattern_inDel.search(x).groups() if x != "Missing" else (nan, '')).apply(Series)
        altSeq_df['before_len'] = altSeq_df.apply(lambda x: len(x['before']) if isinstance(x['before'], str) else len(range(x['begin'], x['end']))+1, axis=1)
        altSeq_df['after_len'] = altSeq_df.after.apply(len)

        sup = DataFrame({
            'Entry': dfrm[dfrm['Alternative sequence'].isnull()].Entry.unique(),
            'ftId': '',
            'begin': nan,
            'end': nan,
            'before': nan,
            'after': nan,
            'description': nan,
            'before_len': nan,
            'after_len': nan,
            'evidence': nan})
        return concat((altSeq_df.drop(columns=['note', 'AltRange', 'AltInfo']), sup), ignore_index=True, sort=False)
    
    @classmethod
    @unsync
    async def fetch_VAR_SEQ_to_DB(cls, accessions, name='VAR_SEQ', **kwargs):
        UniProtAPI.params['columns'] = kwargs.get('columns', 'id,feature(ALTERNATIVE%20SEQUENCE)')
        UniProtAPI.params['from'] = kwargs.get('from', 'ACC+ID')
        UniProtAPI.with_name_suffix = kwargs.get('with_name_suffix', True)
        for res in UniProtAPI.retrieve(accessions, name, **kwargs):
            dfrm = await a_read_csv(res, sep='\t').then(cls.deal_with_alternative_seq)
            if dfrm is None:
                continue
            await cls.sqlite_api.async_insert(cls.sqlite_api.VAR_SEQ, dfrm.to_dict('records'))
    
    @classmethod
    @unsync
    async def fetch_VAR_SEQ_from_DB(cls, accessions, name='VAR_SEQ', **kwargs):
        task = cls.tasks.get(('fetch_VAR_SEQ_from_DB', tuple(accessions)), None)
        if task is not None:
            return task
        res = await cls.sqlite_api.VAR_SEQ.objects.filter(Entry__in=accessions).all()
        if len(res) == 0:
            await cls.fetch_VAR_SEQ_to_DB(accessions, name, **kwargs)
            dfrm = DataFrame(await cls.sqlite_api.VAR_SEQ.objects.filter(Entry__in=accessions).all())
        else:
            dfrm = DataFrame(res)
            rest = tuple(frozenset(accessions) - frozenset(dfrm.Entry))
            if len(rest) > 0:
                await cls.fetch_VAR_SEQ_to_DB(rest, name, **kwargs)
                rest_df = DataFrame(await cls.sqlite_api.VAR_SEQ.objects.filter(Entry__in=rest).all())
                dfrm = concat((dfrm, rest_df), ignore_index=True, sort=False)
        cls.register_task(('fetch_VAR_SEQ_from_DB', tuple(accessions)), dfrm)
        return dfrm
        
        
    



# @Created Date: 2020-08-12 11:27:46 pm
# @Filename: __init__.py
# @Email:  1730416009@stu.suda.edu.cn
# @Author: ZeFeng Zhu
# @Last Modified: 2020-09-26 04:27:17 pm
# @Copyright (c) 2020 MinghuiGroup, Soochow University
from pdb_profiling.processors.pdbe.record import (
    Base, 
    PDB, 
    PDBAssemble,
    PDBInterface,
    SIFTS,
    Compounds,
    PDBs,
    SIFTSs
    )
from pdb_profiling.processors.pdbe.api import PDBeModelServer, PDBArchive, PDBVersioned
from pdb_profiling.processors.uniprot.api import UniProtFASTA, UniProtAPI
from pdb_profiling.processors.uniprot.record import UniProts
from pdb_profiling.processors.proteins.api import ProteinsAPI
from pdb_profiling.processors.proteins.record import Identifier, Identifiers
from pdb_profiling.processors.ensembl.api import EnsemblAPI
from pdb_profiling.processors.eutils.api import EutilsAPI
from pdb_profiling.processors.swissmodel.api import SMR



from pdb_profiling import default_config
from tqdm import tqdm

default_config()

def test_init():
    from pdb_profiling.processors.i3d.api import Interactome3D
    Interactome3D.pipe_init_interaction_meta().result()

def test_single_select():
    from pdb_profiling.processors import SIFTS
    # SIFTS.chain_filter, SIFTS.entry_filter = '', ''
    demo = SIFTS('P21359')
    demo.pipe_select_mo().result()
    demo.pipe_select_ho(run_as_completed=True, progress_bar=tqdm).result()
    demo.pipe_select_he(run_as_completed=True, progress_bar=tqdm).result()

def test_identifiers():
    from pdb_profiling.processors import Identifiers
    demo = Identifiers([
        'ENSP00000491589', 'ENST00000379268',
        'ENSP00000427757', 'ENSP00000266732',
        'NP_001291289.1', 'ENST00000335295',
        'NP_001165602.1', 'ENST00000402254',
        'ENST00000371100'])
    demo.fetch('map2unp').run().result()

def test_uniprots_alt():
    from pdb_profiling.processors import UniProts, Identifiers
    from pdb_profiling.utils import a_concat
    UniProts.fetch_VAR_SEQ_from_DB(('Q5VST9', 'Q5JWF2', 'P08631', 'O92972')).result()
    
    demo_unps = ('Q5VST9', 'Q5JWF2', 'P21359', 'P68871', 'P63092')
    Identifiers(demo_unps).query_from_DB_with_unps('ALTERNATIVE_PRODUCTS').run().then(a_concat).result()



# @Created Date: 2020-02-16 09:58:31 pm
# @Filename: log.py
# @Email:  1730416009@stu.suda.edu.cn
# @Author: ZeFeng Zhu
# @Last Modified: 2020-02-16 09:58:41 pm
# @Copyright (c) 2020 MinghuiGroup, Soochow University
import logging
from typing import Optional, Union
from pathlib import Path

LOG_FORMAT = "%(asctime)s %(name)s %(levelname)s %(message)s"

# logging.basicConfig(format=LOG_FORMAT)

class Abclog(object):
    '''
    Abstract class of logging template
    '''
    @classmethod
    def init_logger(cls, logName: Optional[str] = None, logger: Optional[logging.Logger] = None, log_level: int = logging.DEBUG, stream_log_level: int = logging.WARNING):
        if hasattr(cls, 'logger') and cls.logger is not None:
            pass
        elif logger is None:
            if logName is None:
                logName = cls.__name__
            cls.logger = logging.getLogger(logName)
            cls.logger.setLevel(log_level)
            cls.streamHandler = logging.StreamHandler()
            cls.streamHandler.setLevel(stream_log_level)
            cls.formatter = logging.Formatter(LOG_FORMAT)
            cls.streamHandler.setFormatter(cls.formatter)
            cls.logger.addHandler(cls.streamHandler)
        else:
            cls.logger = logger

    @classmethod
    def set_logging_fileHandler(cls, path: Union[str, Path], level: int = logging.DEBUG, logName: Optional[str] = None):
        if not hasattr(cls, 'logger'):
            cls.init_logger(logName)
        try:
            if path is None:
                return cls.logger
            fileHandler = logging.FileHandler(filename=path)
            fileHandler.setLevel(level)
            fileHandler.setFormatter(cls.formatter)
            cls.logger.addHandler(fileHandler)
            cls.logger.info(f"Logging file in {path}")
            return cls.logger
        except Exception:
            cls.logger.exception("Invalid file path for logging file ! Please specifiy path=...")


class SimpleLog(object):
    
    @classmethod
    def init_logger(cls):
        cls.logger = logging.getLogger(cls.__name__)
        cls.logger.setLevel(logging.DEBUG)



# @Created Date: 2020-09-17 12:02:37 am
# @Filename: api.py
# @Email:  1730416009@stu.suda.edu.cn
# @Author: ZeFeng Zhu
# @Last Modified: 2020-09-17 12:02:45 am
# @Copyright (c) 2020 MinghuiGroup, Soochow University
from typing import Union, Optional, Iterator, Iterable, Set, Dict, List, Any, Generator, Callable, Tuple
from pathlib import Path
from numpy import nan
from pandas import DataFrame, Series
from pdb_profiling.log import Abclog
from pdb_profiling.fetcher.webfetch import UnsyncFetch
from pdb_profiling.utils import dumpsParams, range_len
from pdb_profiling.processors.uniprot.record import UniProts
from urllib.parse import quote
from slugify import slugify
from unsync import unsync


BASE_URL = 'https://www.ebi.ac.uk/proteins/api/'


class ProteinsAPI(Abclog):
    '''
    Implement The Proteins REST API

    * <https://www.ebi.ac.uk/proteins/api/doc/index.html>
    '''
    
    headers = {'Accept': 'application/json'}
    api_set = frozenset((
        'proteins', 'proteins/covid-19/entries', 'proteins/interaction/', 'proteins/',
        'coordinates', 'coordinates/', 'coordinates/location/',
        'uniparc', 'uniparc/accession/', 'uniparc/best/guess',
        'uniparc/dbreference/', 'uniparc/proteome/', 'uniparc/sequence',  # NOTE: uniparc/sequence use POST method!
        'uniparc/upi/', 'variation/dbsnp/', 'variation/hgvs/', 'variation/'))
    
    @classmethod
    def get_file_suffix(cls) -> str:
        res = cls.headers["Accept"].split('/')[1]
        assert res in ('json', 'xml', 'x-gff'), f"Unexpected Case: {cls.headers}"
        return res

    @classmethod
    def task_unit(cls, suffix: str, params: Dict, folder: Path, identifier:Optional[str]=None) -> Tuple:
        args = dict(
            url=f'{BASE_URL}{suffix}' if identifier is None else f'{BASE_URL}{suffix}{quote(identifier)}',
            headers=cls.headers,
            params=params)
        return 'get', args, folder/f'{slugify(identifier)+"_"+dumpsParams(params) if identifier is not None else dumpsParams(params)}.{cls.get_file_suffix()}'

    @classmethod
    def yieldTasks(cls, suffix: str, params_collection: Iterable[Dict], folder: Path, identifiers: Optional[Iterable[str]]) -> Generator:
        # https://www.ebi.ac.uk/proteins/api/coordinates?offset=0&size=-1&ensembl=ENST00000554444
        if identifiers is None:
            for params in params_collection:
                yield cls.task_unit(suffix, params, folder)
        else:
            for identifier, params in zip(identifiers, params_collection):
                yield cls.task_unit(suffix, params, folder, identifier)

    @classmethod
    def retrieve(cls, suffix: str, params_collection: Iterable[Dict], folder: Union[Path, str], identifiers: Optional[Iterable[str]] = None, concur_req: int = 20, rate: float = 1.5, ret_res: bool = True, **kwargs):
        assert suffix in cls.api_set, f"Invalid suffix! Valid set is \n{cls.api_set}"
        folder = Path(folder)
        res = UnsyncFetch.multi_tasks(
            cls.yieldTasks(suffix, params_collection, folder, identifiers),
            concur_req=concur_req,
            rate=rate,
            ret_res=ret_res,
            semaphore=kwargs.get('semaphore', None))
        return res
    
    @classmethod
    def single_retrieve(cls, suffix: str, params:Dict, folder: Union[Path, str], semaphore, identifier:Optional[str]=None, rate: float = 1.5):
        assert suffix in cls.api_set, f"Invalid suffix! Valid set is \n{cls.api_set}"
        folder = Path(folder)
        return UnsyncFetch.single_task(
            task=cls.task_unit(suffix, params, folder, identifier),
            semaphore=semaphore,
            rate=rate)

    @classmethod
    def query_sequence(cls, params: Dict, data: Dict, folder: Union[Path, str], fileName:str, semaphore, rate: float = 1.5):
        '''
        Implement `uniparc/sequence`
        '''
        folder = Path(folder)
        args = dict(
            url=f'{BASE_URL}uniparc/sequence',
            headers=cls.headers,
            params=params,
            data=data)
        return UnsyncFetch.single_task(
            task=('post', args, folder/f'{fileName}.{cls.get_file_suffix()}'),
            semaphore=semaphore,
            rate=rate)
        
    @classmethod
    def flat_data(cls, data, surface=False, except_keys=[]):
        if isinstance(data, dict):
            if len(data) == 1 and 'value' in data:
                return cls.flat_data(data['value'], except_keys=except_keys)
            else:
                new = dict()
                for key, value in data.items():
                    if key in except_keys:
                        new[key] = value
                    else:
                        new[key] = cls.flat_data(value, except_keys=except_keys)
                return new
        elif isinstance(data, list):
            if len(data) == 1 and not surface:
                return cls.flat_data(data[0], except_keys=except_keys)
            else:
                return [cls.flat_data(i, except_keys=except_keys) for i in data]
        else:
            return data

    @classmethod
    @unsync
    async def pipe_summary(cls, data: List):
        if len(data) > 1:
            cls.logger.warning(f"Unexpected Length from ProteinsAPI.pipe_summary: {len(data)}, {[data[i]['accession'] for i in range(len(data))]}")
            '''
            Special Cases Like: P04745 (AMY1A_HUMAN)
                                P0DTE7 (AMY1B_HUMAN)
                                P0DTE8 (AMY1C_HUMAN)
                                with identical sequence
                                Causion from UniProtlKB: 'Three distinct genes (AMY1A, AMY1B and AMY1C), located in a gene cluster on 1p21, encode proteins sharing the same peptidic sequence.'
            '''
            # raise AssertionError("With Unexpected length!")
        elif len(data) == 0:
            return
        data = data[0]
        dbReferences_lyst = []
        common_cols = ('type', 'isoform')
        for i in data['dbReferences']:
            if i['type'] == 'RefSeq':
                dbReferences_lyst.append({
                    **{key: i.get(key, nan) for key in common_cols},
                    **dict(accession=data['accession'], protein=i['id'], transcript=i['properties']['nucleotide sequence ID'], gene=nan),
                })
            elif i['type'] == 'Ensembl':
                dbReferences_lyst.append({
                    **{key: i.get(key, nan) for key in common_cols},
                    **dict(accession=data['accession'], protein=i['properties']['protein sequence ID'], transcript=i['id'], gene=i['properties']['gene ID'])})
        dbReferences_df = DataFrame(dbReferences_lyst)
        other_dbReferences_df = DataFrame((i for i in data['dbReferences'] if i['type'] not in ('RefSeq', 'Ensembl')))
        other_dbReferences_df['accession'] = data['accession']
        if 'isoform' not in other_dbReferences_df.columns:
            other_dbReferences_df['isoform'] = ''
        else:
            other_dbReferences_df['isoform'].fillna('', inplace=True)
        features_df = DataFrame(data['features']); features_df['accession'] = data['accession']
        if 'molecule' not in features_df.columns:
            features_df['molecule'] = ''
        iso_df, int_df = None, None
        if 'comments' in data.keys():
            for comment in data['comments']:
                if comment['type'] == 'ALTERNATIVE_PRODUCTS':
                    iso_df = DataFrame(
                        cls.flat_data(comment['isoforms'], surface=True, except_keys={'sequence'}))
                    iso_df['accession'] = data['accession']
                    iso_df.rename(columns={'ids': 'isoform'}, inplace=True)
                    iso_df['else_iso'] = iso_df.isoform.apply(lambda x: x[1:] if isinstance(x, List) else nan)
                    focus_index = iso_df[(~iso_df.else_iso.isnull())].index
                    iso_df.loc[focus_index, 'isoform'] = iso_df.loc[focus_index, 'isoform'].apply(lambda x: x[0])
                    iso_df.rename(columns={'sequence': 'VAR_SEQ'}, inplace=True)
                    alt_seq_df = await UniProts.fetch_VAR_SEQ_from_DB((data['accession'], ), "%s+VAR_SEQ" % data['accession'], with_name_suffix=False)
                    
                    ac_len = data['sequence']['length']
                    ac_seq = data['sequence']['sequence']
                    
                    if len(alt_seq_df) > 0:
                        assert all(alt_seq_df.ftId.ne('')), f"{alt_seq_df[alt_seq_df.ftId.eq('')]}"
                        alt_seq_dict = alt_seq_df[["ftId", "before_len", "after_len", "after", "begin", "end"]].to_dict('list')
                        iso_alt_info = iso_df.VAR_SEQ.apply(lambda x: UniProts.get_alt_interval_base(x, alt_seq_dict) if not isinstance(x, float) else x)
                        iso_df[["iso_range", "length", "sequence"]] = iso_alt_info.apply(lambda x: UniProts.get_affected_interval(x, ac_len, ac_seq)).apply(Series)
                    else:
                        iso_df['iso_range'] = nan
                        iso_df['length'] = nan
                        iso_df['sequence'] = nan
                    iso_df['iso_range_len'] = iso_df.iso_range.apply(range_len)
                    iso_df.loc[iso_df.sequenceStatus.eq('displayed').idxmax(), ["length", "sequence"]] = (ac_len, ac_seq)

                if comment['type'] == 'INTERACTION':
                    int_df = DataFrame(comment['interactions'])
                    if 'accession1' not in int_df.columns:
                        int_df['accession1'] = data['accession']
                    else:
                        int_df['accession1'].fillna(data['accession'], inplace=True)
                    assert int_df[['accession1', 'accession2', 'interactor1', 'interactor2']].isnull(
                    ).sum().sum() == 0, int_df[int_df.accession2.isnull() | int_df.interactor1.isnull() | int_df.interactor2.isnull()]
                    int_df_cols = frozenset(int_df.columns)
                    assert int_df_cols <= {'accession1', 'accession2', 'gene', 'interactor1',
                                           'interactor2', 'organismDiffer', 'experiments', 'chain1', 'chain2'}, int_df_cols
                    int_df['accession'] = data['accession']

        info_data = {key: data.get(key, nan) for key in (
            'accession', 'id', 'proteinExistence', 'info', 'organism', 'secondaryAccession', 'protein', 'gene')}
        info_data['sequence'] = data['sequence']['sequence']
        info_data['length'] = data['sequence']['length']

        return dbReferences_df, other_dbReferences_df, iso_df, features_df, int_df, info_data



# @Created Date: 2020-08-27 10:35:34 am
# @Filename: api.py
# @Email:  1730416009@stu.suda.edu.cn
# @Author: ZeFeng Zhu
# @Last Modified: 2020-08-27 10:35:41 am
# @Copyright (c) 2020 MinghuiGroup, Soochow University
from pdb_profiling.fetcher.webfetch import UnsyncFetch
from pdb_profiling.utils import pipe_out, init_folder_from_suffix, init_semaphore, a_read_csv, dumpsParams
from pdb_profiling.processors.transformer import Dict2Tabular
from typing import Union, Dict, Generator, Set, Any, Optional, List
from pathlib import Path
from pdb_profiling.log import Abclog
from unsync import unsync
from aiofiles import open as aiofiles_open
import orjson as json
from pandas import Series, concat, isna

BASE_URL: str = 'https://swissmodel.expasy.org/'


class SMR(Abclog):
    '''
    Implement SWISS-MODEL Repository API

        * <https://swissmodel.expasy.org/docs/smr_openapi>
        * <https://swissmodel.expasy.org/docs/repository_help#smr_api>
    
    >>> SMR.retrieve(
        ('Q6NZ36', 'P12755'), 
        init_folder_from_suffix(yourfolder, 'swissmodel/repository/uniprot/'))
    '''

    root = 'repository/uniprot/'
    headers = {'accept': 'text/plain'}
    use_existing: bool = True

    @classmethod
    def set_folder(cls, folder: Union[Path, str]):
        cls.folder = init_folder_from_suffix(Path(folder), 'swissmodel/repository/uniprot/')

    @classmethod
    @unsync
    async def set_web_semaphore(cls, web_semaphore_values):
        cls.web_semaphore = await init_semaphore(web_semaphore_values)

    @staticmethod
    def yieldSMR(data: Dict):
        cols = ('sequence_length',
                'ac', 'id', 'isoid')
        uniprot_entries = data['result']['uniprot_entries']

        assert len(
            uniprot_entries) == 1, f"Unexpected length of uniprot_entries: {uniprot_entries}"

        for col in ('ac', 'id', 'isoid'):
            data['result'][col] = uniprot_entries[0].get(col, None)
        
        for val in data['result']['structures']:
            keys = tuple(val.keys())
            for key in keys:
                if isinstance(val[key], (List, Dict)):
                    val[key] = json.dumps(val[key]).decode('utf-8')

        yield data['result']['structures'], cols, tuple(data['result'][col] for col in cols)

    @classmethod
    def task_unit(cls, unp, params, file_format, folder):
        args = dict(
            url=f'{BASE_URL}{cls.root}{unp}.{file_format}',
            headers=cls.headers,
            params=params)
        return 'get', args, Path(folder)/f'{unp}_{dumpsParams(params)}.{file_format}'

    @classmethod
    def yieldTasks(cls, unps, params: Dict, file_format: str, folder: Union[Path, str]) -> Generator:
        for unp in unps:
            yield cls.task_unit(unp, params, file_format, folder)

    @classmethod
    def retrieve(cls, unps, folder: Optional[Union[Path, str]]=None, params: Dict = dict(provider='swissmodel'), concur_req: int = 20, rate: float = 1.5, file_format: str = 'json', ret_res: bool = True, **kwargs):
        assert file_format in ('json', 'pdb'), "Invalid file format"
        res = UnsyncFetch.multi_tasks(
            cls.yieldTasks(unps, params, file_format, cls.folder if folder is None else folder),
            cls.process,
            concur_req=concur_req,
            rate=rate,
            ret_res=ret_res,
            semaphore=kwargs.get('semaphore', cls.web_semaphore))
        return res

    @classmethod
    def single_retrieve(cls, unp: str, folder: Optional[Union[Path, str]]=None, semaphore=None, params: Dict = dict(provider='swissmodel'), rate: float = 1.5, file_format: str = 'json'):
        assert file_format in ('json', 'pdb'), "Invalid file format"

        return UnsyncFetch.single_task(
            task=cls.task_unit(unp, params, file_format, cls.folder if folder is None else folder),
            semaphore=cls.web_semaphore if semaphore is None else semaphore,
            to_do_func=cls.process,
            rate=rate)

    @classmethod
    @unsync
    async def process(cls, path):
        cls.logger.debug('Start to decode SMR JSON')
        if not isinstance(path, (str, Path)):
            path = await path
        if path is None or str(path).endswith('.pdb'):
            return path
        new_path = Path(str(path).replace('.json', '.tsv'))
        if new_path.exists() and cls.use_existing and (new_path.stat().st_size > 0):
            return new_path
        async with aiofiles_open(path) as inFile:
            try:
                data = json.loads(await inFile.read())
            except Exception as e:
                cls.logger.error(f"Error in {path}")
                raise e
        res = Dict2Tabular.pyexcel_io(cls.yieldSMR(data))
        if res is not None:
            if isinstance(res, Generator):
                count = 0
                for r in res:
                    if r is not None:
                        await pipe_out(df=r, path=new_path, format='tsv', mode='a' if count else 'w')
                        count += 1
                if not count:
                    cls.logger.debug(f"Without Expected Data (swissmodel/repository/uniprot/): {data}")
                    return None
            else:
                await pipe_out(df=res, path=new_path, format='tsv', mode='w')
            cls.logger.debug(f'Decoded file in {new_path}')
            return new_path
        else:
            cls.logger.warning(f"Without Expected Data (swissmodel/repository/uniprot/): {data}")
            return None

    @classmethod
    @unsync
    async def to_dataframe(cls, path):
        path = await path
        if path is None:
            return None
        df = await a_read_csv(path, sep="\t", keep_default_na=False, na_values=['NULL', 'null', ''])
        df.rename(columns={'ac': 'UniProt', 'sequence_length': 'unp_len', 'id': 'identifier'}, inplace=True)
        df['unp_range'] = df.apply(lambda x: f"[[{x['from']},{x['to']}]]", axis=1)
        if 'identity' in df.columns:
            df.identity = df.identity/100
        if 'qmean' in df.columns:
            return concat((
                df.drop(columns=['from', 'to', 'qmean']),
                df.qmean.apply(lambda x: json.loads(x) if not isna(x) else default_qmean).apply(Series)), axis=1)
            # add_prefix('qmean.')
        else:
            return df.drop(columns=['from', 'to'])


default_qmean = {
    "acc_agreement_norm_score": None,
    "acc_agreement_z_score": None,
    "avg_local_score": None,
    "avg_local_score_error": None,
    "cbeta_norm_score": None,
    "cbeta_z_score": None,
    "interaction_norm_score": None,
    "interaction_z_score": None,
    "packing_norm_score": None,
    "packing_z_score": None,
    "qmean4_norm_score": None,
    "qmean4_z_score": None,
    "qmean6_norm_score": None,
    "qmean6_z_score": None,
    "ss_agreement_norm_score": None,
    "ss_agreement_z_score": None,
    "torsion_norm_score": None,
    "torsion_z_score": None
}



# @Created Date: 2020-05-13 08:54:03 pm
# @Filename: __init__.py
# @Email:  1730416009@stu.suda.edu.cn
# @Author: ZeFeng Zhu
# @Last Modified: 2020-05-13 08:54:09 pm
# @Copyright (c) 2020 MinghuiGroup, Soochow University
__version__ = '0.2.1'


def default_config(folder='./'):
    from pdb_profiling.log import Abclog
    from pdb_profiling.fetcher.webfetch import UnsyncFetch
    from pdb_profiling.processors.pdbe.record import Base, PDB
    from pdb_profiling.processors.pdbe.api import ProcessPDBe
    from pdb_profiling.processors.proteins.record import Identifier
    from pdb_profiling.processors.uniprot.api import UniProtFASTA, UniProtAPI
    from pdb_profiling.processors.uniprot.record import UniProts
    from pdb_profiling.processors.i3d.api import Interactome3D
    from pdb_profiling.processors.swissmodel.api import SMR
    # Use Existing Handled PDBe API Results (e.g. tsv format results)
    ProcessPDBe.use_existing = True
    # Use Existing API Results (e.g. json format results downloaded from web)
    UnsyncFetch.use_existing = True
    # Init Abclog Logger
    Abclog.init_logger(logName='PDB-Profiling')
    # Set WebFetcher's Semaphore
    Base.set_web_semaphore(50).result()
    Identifier.set_web_semaphore(30).result()
    UniProtFASTA.set_web_semaphore(30).result()
    UniProtAPI.set_web_semaphore(30).result()
    Interactome3D.set_web_semaphore(30).result()
    SMR.set_web_semaphore(30).result()
    # Set Folder that store downloaded and handled files
    Base.set_folder(folder)
    PDB.set_folder(folder)
    Identifier.set_folder(folder)
    UniProtFASTA.set_folder(folder)
    UniProtAPI.set_folder(folder)
    UniProts.set_folder(folder)
    Interactome3D.set_folder(folder)
    SMR.set_folder(folder)



# @Created Date: 2019-11-20 06:33:41 pm
# @Filename: utils.py
# @Email:  1730416009@stu.suda.edu.cn
# @Author: ZeFeng Zhu
# @Last Modified: 2020-02-17 10:25:37 am
# @Copyright (c) 2020 MinghuiGroup, Soochow University
import os
import gzip
import shutil
from typing import Optional, Union, Dict, Tuple, Iterable, Iterator, List, Coroutine, NamedTuple, Callable, Generator
from logging import Logger
from pandas import read_csv, DataFrame, isna, Series, concat
import numpy as np
from pathlib import Path
from aiofiles import open as aiofiles_open
from re import compile as re_compile
from tablib import Dataset
from pyexcel import Sheet
import asyncio
from unsync import unsync, Unfuture
from itertools import chain
import orjson as json
import logging
from io import StringIO
from textdistance import overlap, sorensen
from scipy.stats import wasserstein_distance
from collections import Counter, OrderedDict


def to_interval(lyst: Union[Iterable, Iterator]) -> List:
    def pass_check(lyst):
        try:
            if (not isinstance(lyst, Generator) and  (len(lyst) == 0)) or isna(lyst):
                return False
            else:
                return True
        except ValueError:
            if isinstance(lyst, float):
                return False
            else:
                return True
    if not pass_check(lyst):
        return ()
    else:
        if not isinstance(lyst, set):
            lyst = frozenset(int(i) for i in lyst if i is not None)
        if not pass_check(lyst):
            return ()
        start = []
        interval_lyst = []
        max_edge = max(lyst)
        min_edge = min(lyst)
        if len(lyst) == (max_edge + 1 - min_edge):
            return ((min_edge, max_edge),)
        else:
            lyst_list = sorted(lyst)
            for j in lyst_list:
                if not start:
                    i = j
                    start.append(j)
                    i += 1
                else:
                    if (i != j) or (j == max(lyst_list)):
                        if j == max(lyst_list):
                            if (i != j):
                                interval_lyst.append(start)
                                interval_lyst.append([j])
                                break
                            else:
                                start.append(j)
                        interval_lyst.append(start)
                        start = [j]
                        i = j + 1
                    else:
                        start.append(j)
                        i += 1
            return tuple((min(li), max(li)) for li in interval_lyst)


def lyst22intervel(x, y):
    # x, y = sorted(x), sorted(y)
    x, y = zip(*sorted(zip(x, y), key=lambda cur: cur[0]))
    start_x, start_y = x[0], y[0]
    index_x, index_y = x[0]-1, y[0]-1
    interval_x, interval_y = [], []
    for i, j in zip(x, y):
        pre_x = index_x + 1
        pre_y = index_y + 1
        if pre_x == i and pre_y == j:
            index_x, index_y = i, j
        else:
            interval_x.append((start_x, index_x))
            interval_y.append((start_y, index_y))
            start_x, start_y = i, j
            index_x, index_y = i, j
    interval_x.append((start_x, index_x))
    interval_y.append((start_y, index_y))
    return interval_x, interval_y


@unsync
async def init_semaphore(concurreq) -> Unfuture:
    """
    `semaphore` initiated in the `unsync` event loop
    """
    await asyncio.sleep(.01)
    return asyncio.Semaphore(concurreq)


def decompression(path: str, extension: str =".gz", remove: bool =True, outputPath: Optional[str] = None, logger: Optional[Logger] = None):
    """
    Decompress gz file

    :param path: The file path of the compressed file
    :param extension: Compressed file tail, default value: `.gz`
    :param remove: Whether remove the compressed file, default value: `True`
    :param outputPath: File safe path, default value: `None`
    :param logger: logger Object
    """

    """
    with gzip.GzipFile(mode="rb", fileobj=open(path, 'rb')) as raw:
        with open(path[:-len(extension)], "wb") as file:
            file.write(raw.read())
    """
    if outputPath is None:
        outputPath = path[:-len(extension)]

    with gzip.open(path, 'rb') as raw:
        with open(outputPath, 'wb') as file:
            shutil.copyfileobj(raw, file)
    try:
        if remove:
            os.remove(path)
    except Exception as e:
        if isinstance(logger, Logger):
            logger.error(e)

    return outputPath


def related_dataframe(filters: Optional[Union[Dict, Iterable[Tuple]]] = None, dfrm: Optional[DataFrame] = None, path: Union[str, Path, None] = None, sep: str = '\t', **kwargs):
    '''
    valid symbol: `eq, ne, le, lt, ge, gt, isin, isnull`
    '''
    
    if dfrm is None:
        if path is not None:
            dfrm = read_csv(path, sep=sep)
        else:
            raise ValueError('path should not be None')
    elif not isinstance(dfrm, DataFrame):
        raise ValueError('dfrm should be a pandas.DataFrame')

    if filters is None:
        return dfrm
    elif isinstance(filters, Dict):
        filters = filters.items()

    for col, (symbol, value) in filters:
        dfrm = dfrm[getattr(getattr(dfrm, col), symbol)(value)]
    return dfrm


def sort_sub_cols(dfrm, cols):
    if set(dfrm.columns) >= set(cols):
        dfrm[cols] = np.sort(dfrm[cols].to_numpy())
        return dfrm.drop_duplicates()
    else:
        return dfrm


@unsync
async def a_read_csv(path, read_mode='r',**kwargs):
    '''
    only suitable for small dataframe
    '''
    try:
        if isinstance(path, (Coroutine, Unfuture)):
            path = await path
        if isinstance(path, (Path, str)):
            async with aiofiles_open(path, read_mode) as file_io:
                with StringIO(await file_io.read()) as text_io:
                    return read_csv(text_io, **kwargs)
        elif isinstance(path, DataFrame):
            return path
        else:
            return DataFrame(path, columns=kwargs.get('columns', None))
    except Exception:
        raise ValueError(f'{path}')


async def a_load_json(path):
    try:
        if isinstance(path, (Coroutine, Unfuture)):
            path = await path
        if path is None:
            return None
        async with aiofiles_open(path) as inFile:
            return json.loads(await inFile.read())
    except Exception:
        raise ValueError(str(path))


async def pipe_out(df, path, **kwargs):
    if not isinstance(df, (DataFrame, Dataset, Sheet)):
        raise TypeError(f"Invalid Object for pipe_out(): {type(df)}")
    if len(df) == 0:
        raise ValueError("Zero record!")
    path = Path(path)
    mode = kwargs.get('mode', 'a')
    clear_headers:bool = path.exists() and mode.startswith('a')
    var_format = kwargs.get('format', 'tsv').lower()
    async with aiofiles_open(path, mode) as file_io:
        if isinstance(df, DataFrame):
            sorted_col = sorted(df.columns)
            if clear_headers:
                headers = None
            else:
                headers = sorted_col
            dataset = Dataset(headers=headers)
            dataset.extend(df[sorted_col].to_records(index=False))
            to_write = dataset.export(var_format, lineterminator='\n')
        elif isinstance(df, Dataset):
            if clear_headers:
                df.headers = None
            to_write = df.export(var_format, lineterminator='\n')
        elif isinstance(df, Sheet):
            df = df.project(sorted(df.colnames))
            if clear_headers:
                df.colnames = []
            to_write = getattr(df, f"get_{var_format}")(lineterminator='\n')
        else:
            pass
        await file_io.write(to_write)


def flatten_dict(data: Dict, root: str, with_root: bool = True):
    if with_root:
        iterator = yield_flatten_dict(data[root], root)
    else:
        iterator = yield_flatten_dict(data[root])
    for key, value in iterator:
        data[key] = value
    del data[root]


def yield_flatten_dict(data: Dict, root: Optional[str] = None):
    if root is None:
        yield from data.items()
    else:
        for key, value in data.items():
            yield f'{root}.{key}', value


def slice_series(se: Iterable) -> Dict:
    '''
    For Sorted Series
    '''
    data = {}
    cur = next(iter(se))
    start = 0
    try:
        for index, i in enumerate(se):
            if i != cur:
                assert cur not in data, "Invalid Series"
                data[cur] = (start, index)
                cur = i
                start = index
        assert cur not in data, "Invalid Series"
        data[cur] = (start, index+1)
    except AssertionError as e:
        logging.error(e)
        raise e
    return data


def split_df_by_chain(df, all_cols, cols_to_split, mode='sep', sep=','):
    '''
    Reference: <https://stackoverflow.com/a/50731258/12876491>
    '''
    def chainer_sep(s):
        return list(chain.from_iterable(s.str.split(sep)))

    def chainer_json(s):
        return list(chain.from_iterable(s.apply(json.loads)))
    
    def chainer_list(s):
        return list(chain.from_iterable(s))
    
    if mode == 'sep':
        chainer = chainer_sep
        lens = df[cols_to_split[0]].str.split(sep).map(len)
    elif mode == 'json-list':
        chainer = chainer_json
        lens = df[cols_to_split[0]].apply(json.loads).map(len)
    elif mode == 'list':
        chainer = chainer_list
        lens = df[cols_to_split[0]].map(len)
    else:
        raise ValueError("Invalid mode!")
    
    repeat_dict = {col: np.repeat(df[col], lens)
                   for col in set(all_cols)-set(cols_to_split)}
    chain_dict = {col: chainer(df[col]) for col in cols_to_split}
    return DataFrame({**repeat_dict, **chain_dict})


def init_folder_from_suffix(folder: Union[Path, str], suffix: str):
    folder = Path(folder)
    new_path = folder/suffix
    new_path.mkdir(parents=True, exist_ok=True)
    return new_path


def init_folder_from_suffixes(folder: Union[Path, str], suffixes: Iterable) -> Iterable[Path]:
    folder = Path(folder)
    for suffix in suffixes:
        new_path = folder/suffix
        new_path.mkdir(parents=True, exist_ok=True)
        yield new_path


def iter_first(df: DataFrame, criteria: Callable[[NamedTuple], bool], **kwargs) -> Optional[NamedTuple]:
    '''
    Implement pandas.DataFrame.itertuples
    
    Returns the value as soon as you find the first row/record 
    that meets the requirements and NOT iterating other rows

    Originated from: https://stackoverflow.com/a/63826677/12876491

    >>> iter_first(df, lambda row: row.A > 4 and row.B > 3)
    '''
    for row in df.itertuples(**kwargs):
        if criteria(row):
            return row


class MMCIF2DictPlus(dict):
    """
    Parse a mmCIF file and return a dictionary
        
    NOTE: Override methods based on Biopython's `Bio.PDB.MMCIF2Dict.MMCIF2Dict`
    """

    def _check_token_with_focus_keys(self, token: Iterable[str]) -> bool:
        is_key, key_value = token
        return is_key == 0 and ((key_value in self.focus_keys) or any(key in key_value for key in self.focus_keys))

    def __init__(self, handle, focus_keys: Iterable[str]=['']):
        self.focus_keys = set(focus_keys)
        self.quote_chars = ["'", '"']
        self.whitespace_chars = [" ", "\t"]
        # TODO: init first loop
        loop_flag = False
        key = None
        tokens = self._tokenize(handle)
        try:
            token = next(tokens)
        except StopIteration:
            return  # NOTE: annotation from biopython: for Python 3.7 and PEP 479
        self[token[1][0:5]] = token[1][5:]
        i = 0
        n = 0
        use = []
        # TODO: loops
        for token in tokens:
            if token[1].lower() == "loop_":
                loop_flag = True
                keys = []
                i = 0
                n = 0
                use = []
                continue
            elif loop_flag:
                '''
                NOTE: annotation from biopython:
                # The second condition checks we are in the first column
                # Some mmCIF files (e.g. 4q9r) have values in later columns
                # starting with an underscore and we don't want to read
                # these as keys
                '''
                if token[1].startswith("_") and (n == 0 or i % n == 0):
                    if i > 0:
                        loop_flag = False
                    else:
                        if self._check_token_with_focus_keys(token):
                            use.append(n)
                            self[token[1]] = []
                        keys.append(token[1])
                        n += 1
                        continue
                else:
                    key_index = i % n
                    try:
                        if key_index in use:
                            self[keys[key_index]].append(token[1])
                    except Exception:
                        raise ValueError(f"{keys}, {key_index}, {use}")
                    i += 1
                    continue
            if key is None:
                if self._check_token_with_focus_keys(token):
                    key = token[1]
            else:
                # Always returns a list
                self[key] = [token[1]]
                key = None
            if self.keys() >= self.focus_keys:
                break

    def _splitline(self, line: str):
        # NOTE: annotation from biopython: See https://www.iucr.org/resources/cif/spec/version1.1/cifsyntax for the syntax
        in_token = False
        # NOTE: annotation from biopython: quote character of the currently open quote, or None if no quote open
        quote_open_char = None
        start_i = 0
        for (i, c) in enumerate(line):
            if c in self.whitespace_chars:
                if in_token and not quote_open_char:
                    in_token = False
                    yield start_i, line[start_i:i]
            elif c in self.quote_chars:
                if not quote_open_char and not in_token:
                    # raise ValueError(f"{self['data_']}: Opening quote in middle of word: " + line)
                    quote_open_char = c
                    in_token = True
                    start_i = i + 1
                elif c == quote_open_char and (i + 1 == len(line) or line[i + 1] in self.whitespace_chars):
                    quote_open_char = None
                    in_token = False
                    yield start_i, line[start_i:i]
            elif c == "#" and not in_token:
                ''' NOTE: annotation from biopython:
                # Skip comments. "#" is a valid non-comment char inside of a
                # quote and inside of an unquoted token (!?!?), so we need to
                # check that the current char is not in a token.
                '''
                return
            elif not in_token:
                in_token = True
                start_i = i
        if in_token:
            yield start_i, line[start_i:]
        if quote_open_char:
            raise ValueError("Line ended with quote open: " + line)

    def _tokenize(self, handle):
        empty = True
        for line in handle:
            empty = False
            if line.startswith("#"):
                continue
            elif line.startswith(";"):
                '''
                NOTE: annotation from biopython:
                # The spec says that leading whitespace on each line must be
                # preserved while trailing whitespace may be stripped.  The
                # trailing newline must be stripped.
                '''
                token_buffer = [line[1:].rstrip()]
                for line in handle:
                    line = line.rstrip()
                    if line.startswith(";"):
                        yield 1, "\n".join(token_buffer)
                        line = line[1:]
                        if line and not line[0] in self.whitespace_chars:
                            raise ValueError("Missing whitespace")
                        break
                    token_buffer.append(line)
                else:
                    raise ValueError("Missing closing semicolon")
            yield from self._splitline(line.strip())
        if empty:
            raise ValueError("Empty file.")


class DisplayPDB(object):

    a_name = 'Asymmetric unit'

    b_name = 'Biological assembly {assembly_id}'

    a_code = 'model-1'

    b_code = 'assembly-{assembly_id}'

    header_unit = '''
                <td>
                    <b>{name}</b> of {pdb_id}
                </td>
    '''

    content_unit = '''
                <td>
                    <img class="display" width="300em" src="https://cdn.rcsb.org/images/structures/{in_code}/{pdb_id}/{pdb_id}_{code}.jpeg"/>
                </td>
    '''

    css = '''
        <style>
            img.display {
                -webkit-filter: invert(1);
                filter: invert(1);
                }
        </style>
    '''

    template = '''
        <table align="center">
            <tr>
            {headers}
            </tr>
            <tr>
            {content}
            </tr>
        </table>
        '''

    @classmethod
    def setting(cls, pdb_id, assemblies):
        headers = [cls.header_unit.format(name=cls.a_name, pdb_id=pdb_id)]
        content = [cls.content_unit.format(pdb_id=pdb_id, in_code=pdb_id[1:3], code=cls.a_code)]
        for assembly_id in assemblies:
            headers.append(cls.header_unit.format(
                name=cls.b_name.format(assembly_id=assembly_id),
                pdb_id=pdb_id))
            content.append(cls.content_unit.format(
                pdb_id=pdb_id, 
                in_code=pdb_id[1:3], 
                code=cls.b_code.format(assembly_id=assembly_id)
            ))
        return ''.join(headers), ''.join(content)

    def show(self, pdb_id, assemblies: Iterable[int]= [1]):
        from IPython.display import display, HTML
        assemblies = sorted(int(i) for i in assemblies if int(i) > 0)
        headers, content = self.setting(pdb_id, assemblies)
        self.table = self.template.format(headers=headers, content=content)
        if self.dark:
            self.table = self.css + self.table
        display(HTML(self.table))
    
    def __init__(self, dark:bool=False):
        self.dark = dark


fasta_pat = re_compile(r'(>.+)\n([A-Z\*\n]+)')
unp_header_pat = re_compile(r'>sp\|(.+)\|')

async def a_seq_reader(path: Union[Unfuture, Union[Path, str]]):
    if isinstance(path, Unfuture):
        path = await path
    async with aiofiles_open(path, 'rt') as handle:
        header, content = fasta_pat.match(await handle.read()).groups()
        content = content.replace('\n', '')
        assert content != '', str(path)
        return header, content


@unsync
async def get_seq_from_parser(res, identifier, seq_only:bool = True):
    async for header, content in await res:
        if (identifier in header) and (f'{identifier}-' not in header):
            return content if seq_only else (header, content)


@unsync
async def get_seqs_from_parser(res, identifiers:Optional[Iterable[str]]=None):
    ret = []
    async for header, content in await res:
        header = unp_header_pat.match(header).group(1)
        if identifiers is None or header in identifiers:
            ret.append((header, content))
    return ret


async def a_seq_parser(path: Union[Unfuture, Coroutine, Path, str]):
    if isinstance(path, (Unfuture, Coroutine)):
        path = await path
    async with aiofiles_open(path, 'rt') as handle:
        header, content = None, ''
        async for line in handle:
            if line.startswith('>'):
                if header is not None:
                    yield header.strip(), content.replace('\n', '')
                header, content = line, ''
            else:
                content += line
        assert header is not None, f"\npath: {path}\ncur_content: {content}"
        yield header.strip(), content.replace('\n', '')

nu2aa_dict = {
    'ATA': 'I', 'ATC': 'I', 'ATT': 'I', 'ATG': 'M',
    'ACA': 'T', 'ACC': 'T', 'ACG': 'T', 'ACT': 'T',
    'AAC': 'N', 'AAT': 'N', 'AAA': 'K', 'AAG': 'K',
    'AGC': 'S', 'AGT': 'S', 'AGA': 'R', 'AGG': 'R',
    'CTA': 'L', 'CTC': 'L', 'CTG': 'L', 'CTT': 'L',
    'CCA': 'P', 'CCC': 'P', 'CCG': 'P', 'CCT': 'P',
    'CAC': 'H', 'CAT': 'H', 'CAA': 'Q', 'CAG': 'Q',
    'CGA': 'R', 'CGC': 'R', 'CGG': 'R', 'CGT': 'R',
    'GTA': 'V', 'GTC': 'V', 'GTG': 'V', 'GTT': 'V',
    'GCA': 'A', 'GCC': 'A', 'GCG': 'A', 'GCT': 'A',
    'GAC': 'D', 'GAT': 'D', 'GAA': 'E', 'GAG': 'E',
    'GGA': 'G', 'GGC': 'G', 'GGG': 'G', 'GGT': 'G',
    'TCA': 'S', 'TCC': 'S', 'TCG': 'S', 'TCT': 'S',
    'TTC': 'F', 'TTT': 'F', 'TTA': 'L', 'TTG': 'L',
    'TAC': 'Y', 'TAT': 'Y', 'TAA': '_', 'TAG': '_',
    'TGC': 'C', 'TGT': 'C', 'TGA': '_', 'TGG': 'W',
}

def translate2aa(seq:str, check:bool=False):
    assert len(seq) % 3 == 0, "Invalid length of dna OR rna sequence!"
    seq = seq.replace('U', 'T')
    p_seq = ""
    for i in range(0, len(seq), 3):
        codon = seq[i:i + 3]
        p_seq += nu2aa_dict[codon]
    if check:
        assert "_" not in p_seq, "Invalid Sequence!"
    return p_seq


def unsync_run(arg):
    @unsync
    async def s_unsync_wrap(arg):
        return await arg 
    return s_unsync_wrap(arg).result()


@unsync
def unsync_wrap(var):
    return var


class SeqRangeReader(object):
    def __init__(self, name_group):
        self.name = name_group  # ('pdb_id', 'chain_id', 'UniProt')
        self.pdb_range = []
        self.unp_range = []

    def output(self):
        if self.pdb_range:
            pdb_range = json.dumps(self.pdb_range).decode('utf-8')
            unp_range = json.dumps(self.unp_range).decode('utf-8')
            return pdb_range, unp_range
        else:
            return self.default_pdb_range, self.default_unp_range

    def check(self, name_group_to_check, data_group):
        self.default_pdb_range = '[[%s, %s]]' % data_group[:2]
        self.default_unp_range = '[[%s, %s]]' % data_group[2:4]

        if self.name == name_group_to_check:
            self.pdb_range.append([int(data_group[0]), int(data_group[1])])
            self.unp_range.append([int(data_group[2]), int(data_group[3])])
        else:
            self.name = name_group_to_check
            self.pdb_range = [[int(data_group[0]), int(data_group[1])]]
            self.unp_range = [[int(data_group[2]), int(data_group[3])]]

        return self.output()


def sort_2_range(unp_range: List, pdb_range: List):
    unp_range, pdb_range = zip(
        *sorted(zip(unp_range, pdb_range), key=lambda x: x[0][0]))
    return unp_range, pdb_range


def flat_dict_in_df(dfrm:DataFrame, targetCol:Union[str, Series], cols:List):
    try:
        new_cols = list(f'{targetCol.name}.{col}' for col in cols)
        dfrm[new_cols] = DataFrame(
            targetCol.apply(lambda x: list(x[col] for col in cols)).to_list(),
            columns=new_cols)
        return dfrm.drop(columns=[targetCol.name])
    except AttributeError:
        assert isinstance(targetCol, str)
        new_cols = list(f'{targetCol}.{col}' for col in cols)
        dfrm[cols] = DataFrame(
            dfrm[targetCol].apply(lambda x: list(x[col] for col in cols)).to_list(),
            columns=new_cols)
        return dfrm.drop(columns=[targetCol])
    

SEQ_DICT = {
    "GLY": "G", "ALA": "A", "SER": "S", "THR": "T", "CYS": "C", "VAL": "V", "LEU": "L",
    "ILE": "I", "MET": "M", "PRO": "P", "PHE": "F", "TYR": "Y", "TRP": "W", "ASP": "D",
    "GLU": "E", "ASN": "N", "GLN": "Q", "HIS": "H", "LYS": "K", "ARG": "R"}

standardAA = list(SEQ_DICT.keys())

standardNu = ['DA', 'DT', 'DC', 'DG', 'DI', 'A', 'U', 'C', 'G', 'I']


def range_len(lyst: Union[List, str, float]) -> int:
    if isinstance(lyst, float) or lyst is None:
        return 0
    elif isinstance(lyst, str):
        lyst = json.loads(lyst)
    length = 0
    for left, right in lyst:
        assert right >= left, f"\n{lyst}"
        length += right - left + 1
    return length


def interval2set(lyst: Union[Iterable, Iterator, str]):
    if isinstance(lyst, str):
        lyst = json.loads(lyst)
    range_set = frozenset()
    for left, right in lyst:
        range_set |= frozenset(range(left, right+1))
    return range_set


def expand_interval(lyst: Union[Iterable, Iterator, str]):
    lyst = json.loads(lyst) if isinstance(lyst, str) else lyst
    yield from (i for start, end in lyst for i in range(start, end+1))


def lyst2range(lyst, add_end=1):
    if isinstance(lyst, str):
        lyst = json.loads(lyst)
    for start, end in lyst:
        yield from range(int(start), int(end)+add_end)


def subtract_range(pdb_range: Union[str, Iterable], mis_range: Union[str, Iterable]) -> List:
    if isinstance(mis_range, float) or mis_range is None:
        return pdb_range
    elif len(pdb_range) == 0:
        return ()
    elif len(mis_range) == 0:
        return pdb_range
    pdb_range_set = interval2set(pdb_range)
    mis_range_set = interval2set(mis_range)
    return to_interval(pdb_range_set - mis_range_set)


def check_range(i) -> bool:
    if isinstance(i, float) or (i is None) or (len(i) == 0) or (i == 'nan'):
        return False
    return True


def add_range(left: Union[str, Iterable], right: Union[str, Iterable]) -> List:
    check_left = check_range(left)
    check_right = check_range(right)
    if check_left and not check_right:
        return left
    elif not check_left and check_right:
        return right
    elif not check_left and not check_right:
        return np.nan
    try:
        left_range_set = interval2set(left)
        right_range_set = interval2set(right)
        return to_interval(left_range_set | right_range_set)
    except Exception as e:
        print(left, right)
        print(type(left), type(right))
        raise e


def overlap_range(obs_range:Union[str, Iterable], unk_range: Union[str, Iterable]) -> List:
    if isinstance(unk_range, float) or unk_range is None:
        return ()
    '''
    obs_range_set = interval2set(obs_range)
    unk_range_set = interval2set(unk_range)
    return to_interval(obs_range_set.intersection(unk_range_set))
    '''
    obs_range = json.loads(obs_range) if isinstance(obs_range, str) else obs_range
    unk_range = json.loads(unk_range) if isinstance(unk_range, str) else unk_range

    def unit(i1,i2):
        for start1, end1 in i1:
            for start2, end2 in i2:
                sl = start2 >= start1
                sr = start2 <= end1
                el = end2 >= start1
                er = end2 <= end1
                s_in = sl and sr
                e_in = el and er
                ini = s_in or e_in
                # out = (sl and el) or (sr and er)
                cov = (not sl) and (not er)
                start = start2 if s_in else start1
                end = end2 if e_in else end1
                if ini or cov:
                    yield start, end

    return tuple(unit(obs_range, unk_range))


def get_seq_seg(seq, ranges):
    for start,end in ranges:
        yield start, seq[start-1:end]


def get_diff_index(lseq, lrange, rseq, rrange):
    if isinstance(lrange, str):
        lrange = json.loads(lrange)
    if isinstance(rrange, str):
        rrange = json.loads(rrange)
    for (lstart, lseg), (rstart, rseg) in zip(get_seq_seg(lseq, lrange), get_seq_seg(rseq, rrange)):
        yield from ((lstart+index, rstart+index) for index, (r1, r2) in enumerate(zip(lseg, rseg)) if r1 != r2)


def red_seq_seg(seq, ranges):
    edge = 0
    for start, end in ranges:
        yield f"{seq[edge:start-1]}\x1b[31m{seq[start-1:end]}\x1b[0m"
        edge = end
    yield seq[end:]


def outside_range(pdb_range: Union[str, Iterable], seqres_len: int):
    pdb_range = json.loads(pdb_range) if isinstance(pdb_range, str) else pdb_range
    out_head = pdb_range[0][0] - 1
    out_tail = pdb_range[-1][-1] + 1
    ret = [[1, out_head], [out_tail, seqres_len]]
    return [i for i in ret if i[0]<=i[1]]


def outside_range_len(pdb_range: Union[str, Iterable], seqres_len: int, omit: int = 5) -> int:
    if isinstance(pdb_range, str):
        lyst = json.loads(pdb_range)
    else:
        lyst = pdb_range
    out_head = lyst[0][0]-1
    out_tail = seqres_len - lyst[-1][-1]
    if out_head <= omit:
        out_head = 0
    else:
        out_head -= omit
    if out_tail <= omit:
        out_tail = 0
    else:
        out_tail -= omit
    return out_head + out_tail


def get_gap_list(li: Union[str,List,Tuple]):
    if isinstance(li, str):
        li = json.loads(li)
    return [li[i+1][0] - li[i][1] - 1 for i in range(len(li)-1)]


def get_range_diff(lyst_a: Union[str, List, Tuple], lyst_b: Union[str, List, Tuple]):
    lyst_a = json.loads(lyst_a) if isinstance(lyst_a, str) else lyst_a
    lyst_b = json.loads(lyst_b) if isinstance(lyst_b, str) else lyst_b
    array_a = np.array([right - left + 1 for left, right in lyst_a])
    array_b = np.array([right - left + 1 for left, right in lyst_b])
    return array_a - array_b


def select_range(ranges, indexes, cutoff=0.2, skip_index=[], selected_ranges=None):
    select_index = []
    selected_ranges = [] if selected_ranges is None else selected_ranges
    def unit(cur_index):
        if cur_index in skip_index:
            return
        cur_range = ranges[cur_index]
        cur_range = json.loads(cur_range) if isinstance(cur_range, str) else cur_range
        for selected_range in selected_ranges:
            selected_range = json.loads(selected_range) if isinstance(selected_range, str) else selected_range
            score = overlap.similarity(lyst2range(cur_range),
                               lyst2range(selected_range))
            if score > cutoff:
                return
        select_index.append(cur_index)
        selected_ranges.append(cur_range)

    for index in indexes:
        unit(index)
    return select_index


def select_ho_range(ranges1, ranges2, indexes, cutoff=0.2, skip_index=[]):
    select_index = []

    def unit(cur_index):
        if cur_index in skip_index:
            return
        cur_range1, cur_range2 = ranges1[cur_index], ranges2[cur_index]
        c1_1 = Counter(expand_interval(cur_range1))
        c1_2 = Counter(expand_interval(cur_range2))
        if len(c1_1) == 0 or len(c1_2) == 0:
            return
        c1 = c1_1 + c1_2
        for selected in select_index:
            selected_range1,selected_range2 = ranges1[selected], ranges2[selected]
            c_c1 = c1.copy()
            c2 = Counter(expand_interval(selected_range1))+Counter(expand_interval(selected_range2))
            for key in c_c1.keys() | c2.keys():
                if key not in c_c1:
                    c_c1[key] = 0
                if key not in c2:
                    c2[key] = 0
            oc2 = OrderedDict(sorted((item for item in c2.items()), key=lambda x: x[0]))
            oc1 = OrderedDict(sorted((item for item in c_c1.items()), key=lambda x: x[0]))
            score = wasserstein_distance(tuple(oc1.values()), tuple(oc2.values()))
            if score < cutoff:
                return
        select_index.append(cur_index)

    for index in indexes:
        unit(index)
    return select_index


def select_he_range(Entry_1, Entry_2, ranges1, ranges2, indexes, cutoff=0.2, skip_index=[]):
    select_index = []

    def unit(cur_index):
        if cur_index in skip_index:
            return
        cur_range1, cur_range2 = ranges1[cur_index], ranges2[cur_index]
        cur_e1, cur_e2 = Entry_1[cur_index], Entry_2[cur_index]
        (cur_e1, cur_range1), (cur_e2, cur_range2) = sorted(((cur_e1, cur_range1), (cur_e2, cur_range2)), key=lambda x: x[0])

        c1_1 = frozenset(f"1_{i}" for i in expand_interval(cur_range1))
        c1_2 = frozenset(f"2_{i}" for i in expand_interval(cur_range2))
        if len(c1_1) == 0 or len(c1_2) == 0:
            return
        c1 = c1_1 | c1_2
        for selected in select_index:
            selected_range1, selected_range2 = ranges1[selected], ranges2[selected]
            selected_e1, selected_e2 = Entry_1[selected], Entry_2[selected]
            (selected_e1, selected_range1), (selected_e2, selected_range2) = sorted(((selected_e1, selected_range1), (selected_e2, selected_range2)), key=lambda x: x[0])
            c2 = frozenset(f"1_{i}" for i in expand_interval(selected_range1)) | frozenset(f"2_{i}" for i in expand_interval(selected_range2))
            score = sorensen.similarity(c1, c2)
            if score > cutoff:
                return
        select_index.append(cur_index)

    for index in indexes:
        unit(index)
    return select_index


def dumpsParams(params: Dict) -> str:
    return '&'.join(f'{key}={value}' for key, value in params.items())


@unsync
async def a_concat(pathes, sep='\t', sort=False, ignore_index=True, columns=None):
    if isinstance(pathes, (Unfuture, Coroutine)):
        pathes = await pathes
    return concat([await a_read_csv((await path) if isinstance(Unfuture, Coroutine) else path, sep=sep, columns=columns) for path in pathes], sort=sort, ignore_index=ignore_index)



# @Created Date: 2020-06-07 06:39:15 pm
# @Filename: api.py
# @Email:  1730416009@stu.suda.edu.cn
# @Author: ZeFeng Zhu
# @Last Modified: 2020-06-07 06:39:23 pm
# @Copyright (c) 2020 MinghuiGroup, Soochow University
from pdb_profiling.utils import sort_sub_cols
from pathlib import Path
from pandas import read_csv, concat
from unsync import unsync
from pdb_profiling.log import Abclog
from pdb_profiling.fetcher.webfetch import UnsyncFetch
from pdb_profiling.utils import init_semaphore, init_folder_from_suffixes, a_read_csv, init_folder_from_suffix
from pdb_profiling.processors.i3d import I3DDB
from typing import Union
from asyncio import as_completed
from pathlib import Path

class I3D_API(object):

    @staticmethod
    def concat_dat(files, outpath):
        outpath = Path(outpath)
        with outpath.open('w+') as out:
            for index, file in enumerate(files):
                with Path(file).open('r+') as cur:
                    if index:
                        next(cur)
                    for line in cur:
                        out.write(line)


    @staticmethod
    def download(folder: Path, orgnisim):
        pass
        

    def __init__(self, path, identity_cutoff: int = 90):
        self.identity_cutoff = identity_cutoff
        self.all = read_csv(path, sep="\t")
        
        self.all_unp_pairs = self.all[['PROT1', 'PROT2']].drop_duplicates().reset_index(drop=True)
        self.all_unp_pairs['from_i3d_unp_pairs'] = True
        self.all_unp_pairs_1 = self.all_unp_pairs
        self.all_unp_pairs_2 = self.all_unp_pairs.rename(columns={"PROT1":"PROT2", "PROT2":"PROT1"})
        self.all_unp_pairs_12 = concat([self.all_unp_pairs_1, self.all_unp_pairs_2]).drop_duplicates()

        self.chain_pairs = self.all[
            self.all.TYPE.eq('Structure') &
            self.all.SEQ_IDENT1.ge(identity_cutoff) &
            self.all.SEQ_IDENT2.ge(identity_cutoff)
        ][['PDB_ID', 'CHAIN1', 'CHAIN2']].drop_duplicates()
        self.chain_pairs = self.pipe_i3d(self.chain_pairs)
        
        self.unp_chain_pairs = self.all[
            self.all.TYPE.eq('Structure') &
            self.all.SEQ_IDENT1.ge(identity_cutoff) &
            self.all.SEQ_IDENT2.ge(identity_cutoff)
        ][['PROT1', 'PROT2', 'PDB_ID', 'CHAIN1', 'CHAIN2']].drop_duplicates()
        self.unp_chain_pairs['partner_1'] = self.unp_chain_pairs.CHAIN1+'_'+self.unp_chain_pairs.PROT1
        self.unp_chain_pairs['partner_2'] = self.unp_chain_pairs.CHAIN2+'_'+self.unp_chain_pairs.PROT2
        self.unp_chain_pairs = self.unp_chain_pairs[['PDB_ID', 'partner_1', 'partner_2']]
        self.unp_chain_pairs = self.pipe_i3d(self.unp_chain_pairs, ['partner_1', 'partner_2'], "from_i3d_unp_chain_pair")
        self.unp_chain_pairs[['chain_id_1', 'PROT1']] = self.unp_chain_pairs.partner_1.str.split('_', expand=True)
        self.unp_chain_pairs[['chain_id_2', 'PROT2']] = self.unp_chain_pairs.partner_2.str.split('_', expand=True)
        self.unp_chain_pairs.drop(columns=['partner_1', 'partner_2'], inplace=True)

    
    @staticmethod
    def pipe_i3d(df, focus_cols=['chain_id_1', 'chain_id_2'], newcol="from_i3d_chain_pair"):
        df = df.rename(columns={
                        "PDB_ID": "pdb_id", 
                        "CHAIN1": "chain_id_1",
                        "CHAIN2": "chain_id_2"}).reset_index(drop=True)
        df[newcol] = True
        return sort_sub_cols(df, focus_cols)
    
    def chain_pairs_p(self, pdbs):
        return self.chain_pairs[self.chain_pairs.pdb_id.isin(pdbs)]
    
    def unp_chain_pairs_p(self, pdbs):
        return self.unp_chain_pairs[self.unp_chain_pairs.pdb_id.isin(pdbs)]


class Interactome3D(Abclog):
    '''
    Implement Interactome3D API
    
    DEMO URL: https://interactome3d.irbbarcelona.org/user_data/bsubtilis/download/complete/interactions.dat
    '''
    root = 'https://interactome3d.irbbarcelona.org/'
    headers = {'accept': 'text/plain'}
    organisms = frozenset(
        {'human', 'athaliana', 'bsubtilis', 'btaurus', 'celegans', 
         'cjejuni', 'fly', 'ecoli', 'hpylori', 'mouse',
         'mtuberculosis', 'mpneumoniae', 'pfalciparum', 
         'rat', 'yeast', 'spombe', 'ssp', 'tpallidum'})
    api_set = frozenset({'api/getProteinStructures', 'api/getInteractionStructures', 'api/getUniprotACs',
                         'api/getPdbFile', 'api/getVersion'})


    @classmethod
    def set_folder(cls, folder: Union[Path, str]):
        folder = Path(folder)
        assert folder.exists(), "Folder not exist! Please create it or input a valid folder!"
        cls.folder, cls.split_folder = tuple(init_folder_from_suffixes(folder, ('interactome3d', 'interactome3d/split')))
        cls.sqlite_api = I3DDB(
            "sqlite:///%s" % (init_folder_from_suffix(cls.folder.parent, 'local_db')/"I3DDB.db"))

    @classmethod
    @unsync
    async def set_web_semaphore(cls, web_semaphore_value):
        cls.web_semaphore = await init_semaphore(web_semaphore_value)

    @classmethod
    def get_web_semaphore(cls):
        return cls.web_semaphore

    @classmethod
    def task_unit(cls, folder, suffix, file_suffix, file_name, params=None):
        args = dict(
            url=f'{cls.root}{suffix}',
            headers=cls.headers)
        if params is not None:
            args['params'] = params
        return 'get', args, folder/f'{file_name}{file_suffix}'

    @classmethod
    def retrieve_metadata(cls, organism: str, dataset: str = 'complete', suffix: str = 'user_data/{organism}/download/{dataset}/', ret_res:bool=True):
        assert organism in cls.organisms, f"Invalid organism!\nValid set:{cls.organisms}"
        res = UnsyncFetch.single_task(cls.task_unit(cls.folder,
            f'{suffix.format(organism=organism, dataset=dataset)}interactions.dat', '.tsv', f'{organism}_{dataset}_interactions'), 
            cls.get_web_semaphore())
        if ret_res:
            return res.result()
        else:
            return res

    @classmethod
    def retrieve_all_meta(cls, dataset: str = 'complete', suffix: str = 'user_data/{organism}/download/{dataset}/'):
        tasks = [cls.task_unit(cls.folder, f'{suffix.format(organism=organism, dataset=dataset)}interactions.dat', '.tsv', f'{organism}_{dataset}_interactions') for organism in cls.organisms]
        return UnsyncFetch.multi_tasks(tasks, semaphore=cls.get_web_semaphore())

    @classmethod
    @unsync
    async def fetch(cls, suffix: str, file_suffix: str = '.xml', **kwargs):
        assert suffix in cls.api_set, f"Invalid api!\nValid set:{cls.api_set}"
        file_name = kwargs.get('filename', False)
        return await UnsyncFetch.single_task(cls.task_unit(
                cls.split_folder,
                suffix, 
                '' if file_name else file_suffix , 
                file_name if file_name else '&'.join(f'{key}={value}' for key, value in kwargs.items()),
                params=kwargs),
            cls.get_web_semaphore())

    @classmethod
    @unsync
    async def reformat(cls, path):
        dfrm = await a_read_csv(path, sep='\t', na_values=['-'])
        dfrm.rename(columns={
            "PROT1": "Entry_1", "PROT2": "Entry_2",
            "BIO_UNIT": "assembly_id", 'PDB_ID': 'pdb_id',
            "CHAIN1": "chain_id_1", "CHAIN2": "chain_id_2",
            "MODEL1": "model_id_1", "MODEL2": "model_id_2"
        }, inplace=True)
        dfrm.assembly_id = dfrm.assembly_id.fillna(0).astype(int)
        dfrm.model_id_1 = dfrm.model_id_1.fillna(0).astype(int)
        dfrm.model_id_2 = dfrm.model_id_2.fillna(0).astype(int)
        dfrm.model_id_1 += 1
        dfrm.model_id_2 += 1
        dfrm = dfrm[dfrm.TYPE.eq('Structure')].drop(columns=[
            'RANK_MAJOR', 'RANK_MINOR', 'TYPE', 'FILENAME',
            'SEQ_IDENT1','COVERAGE1','SEQ_BEGIN1','SEQ_END1','DOMAIN1',
            'SEQ_IDENT2','COVERAGE2','SEQ_BEGIN2','SEQ_END2','DOMAIN2']).reset_index(drop=True)
        assert dfrm.shape[0] == dfrm.drop_duplicates().shape[0]
        return dfrm

    @classmethod
    @unsync
    async def pipe_init_interaction_meta(cls, progress_bar=None):
        @unsync
        async def task_unit(org):
            df = await cls.retrieve_metadata(org, ret_res=False).then(cls.reformat)
            df['organism'] = org
            df['interaction_type'] = df.apply(lambda x: 'ho' if x['Entry_1'] == x['Entry_2'] else 'he', axis=1)
            cls.sqlite_api.sync_insert(cls.sqlite_api.InteractionMeta, df.to_dict('records'))

        tasks = [task_unit(org) for org in cls.organisms]
        if progress_bar is None:
            [await task for task in as_completed(tasks)]
        else:
            [await task for task in progress_bar(as_completed(tasks), total=len(tasks))]
            



# @Created Date: 2020-09-26 01:54:56 pm
# @Filename: api.py
# @Email:  1730416009@stu.suda.edu.cn
# @Author: ZeFeng Zhu
# @Last Modified: 2020-09-26 01:55:04 pm
# @Copyright (c) 2020 MinghuiGroup, Soochow University
from typing import Dict, Tuple, Iterable, Generator, Union, Optional
from pathlib import Path
from pdb_profiling.log import Abclog
from pdb_profiling.fetcher.webfetch import UnsyncFetch


BASE_URL = 'https://rest.ensembl.org/'


class EnsemblAPI(Abclog):
    '''
    Implement The Ensembl REST API

    * <https://rest.ensembl.org/documentation/>
    * DEMO: <https://rest.ensembl.org/sequence/id/ENST00000288602?content-type=text/x-fasta;type=protein>
    '''
    headers = {"Content-Type": "text/x-fasta"}
    api_set = frozenset(('sequence/id/', 'archive/id/'))

    @classmethod
    def get_file_suffix(cls, headers: Optional[Dict]) -> str:
        res = headers["Content-Type"].split('/')[1]
        assert res in ('plain', 'x-seqxml+xml', 'x-fasta', 'json'), f"Unexpected Case: {cls.headers}"
        return res.replace('x-', '').replace('seqxml+', '')

    @classmethod
    def task_unit(cls, suffix: str, identifier: str, params: Optional[Dict], folder: Path, headers:Optional[Dict]) -> Tuple:
        headers = cls.headers if headers is None else headers
        if params is not None:
            args = dict(
                url=f'{BASE_URL}{suffix}{identifier}',
                headers=headers,
                params=params)
        else:
            args = dict(
                url=f'{BASE_URL}{suffix}{identifier}',
                headers=headers)
        return 'get', args, folder/f'{identifier}.{cls.get_file_suffix(headers)}'

    @classmethod
    def yieldTasks(cls, suffix: str, identifiers: Iterable[str], params_collection: Iterable[Optional[Dict]], folder: Path, headers: Optional[Dict]) -> Generator:
        for identifier, params in zip(identifiers, params_collection):
            yield cls.task_unit(suffix, identifier, params, folder, headers)
    
    @classmethod
    def retrieve(cls, suffix: str, identifiers: Iterable[str], params_collection: Iterable[Optional[Dict]], folder: Union[Path, str], concur_req: int = 20, rate: float = 1.5, ret_res: bool = True, headers: Optional[Dict] = None, **kwargs):
        assert suffix in cls.api_set, f"Invalid suffix! Valid set is \n{cls.api_set}"
        folder = Path(folder)
        res = UnsyncFetch.multi_tasks(
            cls.yieldTasks(suffix, identifiers, params_collection, folder, headers),
            concur_req=concur_req,
            rate=rate,
            ret_res=ret_res,
            semaphore=kwargs.get('semaphore', None))
        return res

    @classmethod
    def single_retrieve(cls, suffix: str, identifier: str, params: Optional[Dict], folder: Union[Path, str], semaphore, rate: float = 1.5, headers: Optional[Dict] = None):
        assert suffix in cls.api_set, f"Invalid suffix! Valid set is \n{cls.api_set}"
        folder = Path(folder)
        return UnsyncFetch.single_task(
            task=cls.task_unit(suffix, identifier, params, folder, headers),
            semaphore=semaphore,
            rate=rate)



# @Created Date: 2020-06-01 07:04:11 pm
# @Filename: custom.py
# @Email:  1730416009@stu.suda.edu.cn
# @Author: ZeFeng Zhu
# @Last Modified: 2020-06-01 07:05:14 pm
# @Copyright (c) 2020 MinghuiGroup, Soochow University
from re import compile as re_compile


pat_hgmd = re_compile(r"(.+)\.[0-9]+:p\.([A-z]{1})([0-9]+)([A-z]{1})")
pat_cgi = re_compile(r"([A-z]{1})([0-9]+)([A-z]{1})")

def handle_hgmd_df(hgmd_df):
    '''
    usecols=['PROT']
    '''
    hgmd_df[['from_id', 'Ref', 'Pos', 'Alt']] = hgmd_df.apply(
        lambda x: pat_hgmd.search(x['PROT']).groups(), axis=1, result_type='expand')
    hgmd_df.drop(columns=['PROT'], inplace=True)
    hgmd_df.Pos = hgmd_df.Pos.astype(int)
    return hgmd_df.drop_duplicates().to_dict('records')


def handle_exac_df(exac_df):
    '''
    usecols=['ENST', 'aa_ref', 'aa_pos', 'aa_alt']
    '''
    columns = dict(
        zip(
            ('ENST', 'aa_ref', 'aa_pos', 'aa_alt'),
            ('from_id', 'Ref', 'Pos', 'Alt')
        )
    )
    exac_df.rename(columns=columns, inplace=True)
    return exac_df.drop_duplicates().to_dict('records')


def handle_premptc(premptc_df):
    def split_df(df, colName, sep):
        return df.drop([colName], axis=1).join(df[colName].str.split(sep, expand=True).stack().reset_index(level=1, drop=True).rename(colName))
    '''
    usecols=['Transcript_ID', 'aaref', 'aapos', 'aaalt']
    '''
    columns = dict(
        zip(
            ('Transcript_ID', 'aaref', 'aapos', 'aaalt'),
            ('from_id', 'Ref', 'Pos', 'Alt')
        )
    )
    premptc_df.rename(columns=columns, inplace=True)
    premptc_df = split_df(premptc_df, 'from_id', ';').drop_duplicates()
    return premptc_df.drop_duplicates().to_dict('records')


def handle_cgi_df(cgi_df):
    '''
    usecols=['gene', 'mutation_unp']
    '''
    cgi_df[['Ref', 'Pos', 'Alt']] = cgi_df.apply(
        lambda x: pat_cgi.search(x['mutation_unp']).groups(), axis=1, result_type='expand')
    cgi_df.rename(columns={'gene': 'from_id'}, inplace=True)
    return cgi_df.drop_duplicates().to_dict('records')



import pytest
import importlib

def test_default_config():
    from pdb_profiling import default_config
    default_config()

def import_third_party_packages() -> bool:
    return importlib.import_module("unsync.unsync") is not None

def import_self_modules():
    return all((
        importlib.import_module("pdb_profiling.fetcher.webfetch") is not None,
        importlib.import_module("pdb_profiling.utils") is not None,
        importlib.import_module("pdb_profiling.processors.pdbe.api") is not None,
        importlib.import_module("pdb_profiling.processors.pdbe.record") is not None))


def test_import():
    assert import_third_party_packages()
    assert import_self_modules()

def test_retrieve():
    from pdb_profiling.processors.pdbe.record import PDB
    assert PDB('1a01').status['status_code'] == 'REL'


# @Created Date: 2020-05-26 08:02:13 pm
# @Filename: sqlite_api.py
# @Email:  1730416009@stu.suda.edu.cn
# @Author: ZeFeng Zhu
# @Last Modified: 2020-05-26 08:02:16 pm
# @Copyright (c) 2020 MinghuiGroup, Soochow University
from __future__ import absolute_import
import asyncio
import databases
import orm
import sqlalchemy
from sqlite3 import OperationalError
from time import perf_counter
import pandas as pd
from unsync import unsync
from typing import Dict, Iterable
import logging
# logging.basicConfig(level=logging.INFO)

converters = {
    "pdb_id": str,
    "entity_id": str,
    "chain_id": str,
    "PDB_REV_DATE_ORIGINAL": str,
    "FIRST_REV_DATE": str,
    "PDB_REV_DATE": str,
    "REVISION_DATE": str,
    "resolution": float,
    "METHOD_CLASS": str,
    "BOUND_LIGAND_COUNT": int,
    "BOUND_MOL_COUNT": int,
    "nucleotides_entity_type": str,
    "has_hybrid_nucleotides": bool,
    # SEQRES_Info
    "SEQRES_COUNT": int,
    "AVG_OBS_RATIO": float,
    "AVG_OBS_OBS_RATIO": float,
    "NON_INDEX": str,
    "UNK_INDEX": str,
    "MIS_INDEX": str,
    "UNK_COUNT": int,
    "PURE_SEQRES_COUNT": int,
    "OBS_RECORD_COUNT": int,
    "OBS_UNK_COUNT": int,
    "ATOM_RECORD_COUNT": int,
    # SIFTS_Info
    "UniProt": str,
    "identity": float,
    "pdb_range": str,
    "unp_range": str,
    "group_info": int,
    "pdb_gap_list": str,
    "unp_gap_list": str,
    "var_list": str,
    "repeated": bool,
    "var_0_count": int,
    "unp_gap_0_count": int,
    "unp_pdb_var": int,
    "sifts_range_tag": str,
    "new_unp_range": str,
    "new_pdb_range": str,
    "conflict_range": str,
    # Res_Info
    "residue_number": int,
    "residue_name": str,
    "obs_ratio": float,
    "author_residue_number": int,
    "author_insertion_code": str,
    # Site_Info
    "from_id": str,
    "Ref": str,
    "Pos": int,
    "Alt": str,
    # Else
    "oligo_state": str,
    "RAW_BS": float
    }


class Sqlite_API(object):

    metadata = sqlalchemy.MetaData()

    def init_table_model(self):
        class Entry_Info(orm.Model):
            __tablename__ = 'entry_info'
            __metadata__ = self.metadata
            __database__ = self.database
            pdb_id = orm.String(max_length=4, primary_key=True)
            PDB_REV_DATE_ORIGINAL = orm.String(max_length=19)
            FIRST_REV_DATE = orm.String(max_length=19)
            PDB_REV_DATE = orm.String(max_length=19)
            REVISION_DATE = orm.String(max_length=19)
            resolution = orm.Float(allow_null=True)
            METHOD_CLASS = orm.String(max_length=50, primary_key=True)
            BOUND_LIGAND_COUNT = orm.Integer()
            BOUND_MOL_COUNT = orm.Integer()
            nucleotides_entity_type = orm.String(
                max_length=100, allow_null=True, allow_blank=True, default='')
            has_hybrid_nucleotides = orm.Boolean()

        class SEQRES_Info(orm.Model):
            __tablename__ = 'seqres_info'
            __metadata__ = self.metadata
            __database__ = self.database
            pdb_id = orm.String(max_length=4, primary_key=True)
            entity_id = orm.String(max_length=4, primary_key=True)
            chain_id = orm.String(max_length=4, primary_key=True)
            SEQRES_COUNT = orm.Integer()
            AVG_OBS_RATIO = orm.Float()
            AVG_OBS_OBS_RATIO = orm.Float()
            NON_INDEX = orm.Text(allow_null=True, allow_blank=True, default='')
            UNK_INDEX = orm.Text(allow_null=True, allow_blank=True, default='')
            MIS_INDEX = orm.Text(allow_null=True, allow_blank=True, default='')
            UNK_COUNT = orm.Integer()
            PURE_SEQRES_COUNT = orm.Integer()
            OBS_RECORD_COUNT = orm.Integer()
            OBS_UNK_COUNT = orm.Integer()
            ATOM_RECORD_COUNT = orm.Integer()

        class SIFTS_Info(orm.Model):
            __tablename__ = 'sifts_info'
            __metadata__ = self.metadata
            __database__ = self.database
            UniProt = orm.String(max_length=20, primary_key=True)
            pdb_id = orm.String(max_length=4, primary_key=True)
            entity_id = orm.String(max_length=4, primary_key=True)
            chain_id = orm.String(max_length=4, primary_key=True)
            identity = orm.Float()
            pdb_range = orm.Text()
            unp_range = orm.Text()
            group_info = orm.Integer()
            pdb_gap_list = orm.Text()
            unp_gap_list = orm.Text()
            var_list = orm.Text()
            repeated = orm.Boolean()
            var_0_count = orm.Integer()
            unp_gap_0_count = orm.Integer()
            unp_pdb_var = orm.Integer()
            sifts_range_tag = orm.String(max_length=25)
            new_unp_range = orm.Text()
            new_pdb_range = orm.Text()
            conflict_range = orm.Text(
                allow_null=True, allow_blank=True, default='')

        class PDBRes_Info(orm.Model):
            '''
            entry.ID as pdb_id, 
            entity.ID as entity_id, 
            chain.AUTH_ASYM_ID as chain_id, 
            res.CHEM_COMP_ID as residue_name, 
            toInteger(res.ID) as residue_number, 
            tofloat(inChain.OBSERVED_RATIO) as obs_ratio, 
            toInteger(inChain.AUTH_SEQ_ID) as author_residue_number, 
            inChain.PDB_INS_CODE as author_insertion_code ORDER BY residue_number
            '''
            __tablename__ = 'pdbres_info'
            __metadata__ = self.metadata
            __database__ = self.database
            pdb_id = orm.String(max_length=4, primary_key=True)
            entity_id = orm.String(max_length=4, primary_key=True)
            chain_id = orm.String(max_length=4, primary_key=True)
            residue_number = orm.Integer(primary_key=True)
            residue_name = orm.String(max_length=10)
            obs_ratio = orm.Float()
            author_residue_number = orm.Integer()
            author_insertion_code = orm.String(
                max_length=4, allow_null=True, allow_blank=True, default='')
            
        class Site_Info(orm.Model):
            __tablename__ = 'site_info'
            __metadata__ = self.metadata
            __database__ = self.database
            from_id = orm.String(max_length=20, primary_key=True)
            Ref = orm.String(max_length=3, primary_key=True)
            Pos = orm.Integer(primary_key=True)
            Alt = orm.String(max_length=3, primary_key=True)
            # code = orm.String(max_length=3, primary_key=True, allow_blank=True, allow_null=True, default='')
    
        self.Entry_Info = Entry_Info
        self.SEQRES_Info = SEQRES_Info
        self.SIFTS_Info = SIFTS_Info
        self.PDBRes_Info = PDBRes_Info
        self.Site_Info = Site_Info

    def __init__(self, url: str, drop_all: bool=False, insert_sleep: float=45.5):
        self.insert_sleep = insert_sleep
        self.database = databases.Database(url)
        self.engine = sqlalchemy.create_engine(url)
        self.engine.execute("PRAGMA journal_mode=WAL")
        self.init_table_model()
        if drop_all:
            self.metadata.drop_all(self.engine, checkfirst=True)
        self.metadata.create_all(self.engine, checkfirst=True)

    def sync_insert(self, table, values: Iterable[Dict], prefix_with: str = "OR IGNORE"):
        self.engine.execute(
            table.__table__.insert().prefix_with(prefix_with),
            values)
    
    @unsync
    async def async_insert(self, table, values: Iterable[Dict], prefix_with: str = "OR IGNORE"):
        '''
        # BAD CODE
        await self.database.execute_many(
            query=table.__table__.insert().prefix_with(prefix_with),
            values=values)
        '''
        while True:
            try:
                query = table.__table__.insert().values(values).prefix_with(prefix_with)
                await self.database.execute(query)
                break
            except OperationalError as e:
                logging.error(f"{e}, sleep {self.insert_sleep}s and try again")
                await asyncio.sleep(self.insert_sleep)


@unsync
async def main():
    start = perf_counter()
    # sqlite_api = Sqlite_API("sqlite:///../../../test/db/orm_db_test.db")
    sqlite_api = Sqlite_API("sqlite:///C:\\Download\\m3d_running_example\\DB\\db2130.db")
    print('init db: {:.5f}s'.format(perf_counter()-start))
    '''
    for index, (Info, file) in enumerate(zip((sqlite_api.Entry_Info, sqlite_api.SEQRES_Info, sqlite_api.SIFTS_Info, sqlite_api.PDBRes_Info), (  #
            "C:/Download/20200525/LHY/entry_info.tsv",
            "C:/Download/20200525/LHY/seqres_info.tsv",
            "C:/Download/20200525/LHY/sifts_mapping.tsv",
            "C:/Download/20200525/LHY/pdbres_info.tsv"))):
        if index:
            start = perf_counter()
            values = pd.read_csv(file, sep="\t",
                             converters=converters).drop_duplicates().to_dict('records')
        else:
            start = perf_counter()
            values = pd.read_csv(file, sep="\t",
                                 converters=converters).drop_duplicates(subset=['pdb_id'], keep='last').to_dict('records')
        print('init read_csv: {:.5f}s'.format(perf_counter()-start))
        # query = Info.__table__.insert().prefix_with("OR IGNORE")
        start = perf_counter()
        # await database.execute_many(query=query, values=values)
        # engine.execute(query, values)
        sqlite_api.sync_insert(Info, values)
        print('init insert: {:.5f}s'.format(perf_counter()-start))
    '''
    '''
    site_df = pd.read_csv(
        r'C:\Download\20200525\LHY\site_info_demo.tsv', sep='\t', converters=converters)
    start = perf_counter()
    sqlite_api.sync_insert(sqlite_api.Site_Info, site_df.to_dict('records'))
    print('init insert: {:.5f}s'.format(perf_counter()-start))
    '''
    start = perf_counter()
    # entries = await SIFTS_Info.objects.all()
    # example = await SIFTS_Info.objects.get(UniProt='Q92793')
    # example = await Entry_Info.objects.filter(pdb_id__in=('4u7t', '6g6j')).all()
    # example = await SEQRES_Info.objects.filter(pdb_id__in=('4u7t', '6g6j')).all()
    # example = await SIFTS_Info.objects.filter(UniProt__in=('P12270', 'Q14980')).all()
    # example = await SIFTS_Info.objects.filter(UniProt='P12270').all()
    # example = await PDBRes_Info.objects.filter(pdb_id='4loe', entity_id="1", chain_id='C', obs_ratio__lt=1).all()
    # filter(pdb_id__in=['3wwq', '5v8w'])
    example = await sqlite_api.Entry_Info.objects.limit(10).all()
    check = await sqlite_api.Site_Info.objects.filter(from_id__contains=";").all()
    # example = await sqlite_api.SEQRES_Info.objects.filter(pdb_id='4h3k',entity_id='3',chain_id='F').all()
    # example = await sqlite_api.PDBRes_Info.objects.filter(obs_ratio__lt=1).distinct()
    # example = await sqlite_api.Site_Info.objects.filter(from_id__in=('ENST00000379410', 'ENST00000379409')).all()
    # .filter(obs_ratio__lt=1).
    print('init select: {:.5f}s'.format(perf_counter()-start))
    # res = pd.DataFrame(example)
    # print(example[0].NON_INDEX)
    print({value.pdb_id: (value.resolution, value.REVISION_DATE)
           for value in example})
    # print(len(res), len(res.drop_duplicates()))
    # sqlite_api.sync_insert(sqlite_api.PDBRes_Info, [])


if __name__ == '__main__':
    main().result()



# @Created Date: 2019-12-08 06:46:49 pm
# @Filename: api.py
# @Email:  1730416009@stu.suda.edu.cn
# @Author: ZeFeng Zhu
# @Last Modified: 2020-02-16 10:54:32 am
# @Copyright (c) 2020 MinghuiGroup, Soochow University
from typing import Iterable, Iterator, Optional, Union, Generator, Dict, List
from time import perf_counter
from numpy import nan, array
from pathlib import Path
from unsync import unsync, Unfuture
from copy import deepcopy
from pdb_profiling.log import Abclog
from pdb_profiling.utils import init_semaphore, init_folder_from_suffix, a_read_csv
from pdb_profiling.fetcher.webfetch import UnsyncFetch
from uuid import uuid4
# import logging
# from collections import Counter
# from pdb_profiling.processors.uniprot.process import ExtractIsoAlt


QUERY_COLUMNS: List[str] = [
    'id', 'length', 'reviewed',
    'comment(ALTERNATIVE%20PRODUCTS)',
    'feature(ALTERNATIVE%20SEQUENCE)',
    'genes', 'organism', 'protein%20names']


RESULT_COLUMNS: List[str] = [
    'Entry', 'Length', 'Status',
    'Alternative products (isoforms)',
    'Alternative sequence',
    'Gene names', 'Organism', 'Protein names']


COLUMNS_DICT: Dict = dict(zip(QUERY_COLUMNS, RESULT_COLUMNS))


RESULT_NEW_COLUMN: List[str] = ['yourlist', 'isomap']


BASE_URL: str = 'https://www.uniprot.org'


PARAMS: Dict = {
    # 'fil': 'organism%3A"Homo+sapiens+(Human)+[9606]"+AND+reviewed%3Ayes',
    # reviewed:yes+AND+organism:9606
    'columns': None,
    'query': None,
    'from': None,
    'to': 'ACC',
    'format': 'tab'}

"""
class MapUniProtID(Abclog):
    '''
    Implement UniProt Retrieve/ID Mapping API
    '''

    def __init__(self, id_col: str, id_type: str,
                 dfrm: Optional[DataFrame],
                 ids: Optional[Iterable] = None,
                 sites: Optional[Iterable] = None,
                 genes: Optional[Iterable] = None,
                 usecols: Optional[Iterable] = QUERY_COLUMNS,
                 site_col: Optional[str] = None,
                 gene_col: Optional[str] = None,
                 logger: Optional[logging.Logger] = None,
                 loggingPath: Optional[str] = None):

        self.init_logger(self.__class__.__name__, logger)
        if dfrm is not None:
            self.dfrm = dfrm.drop_duplicates().reset_index(drop=True)
        else:
            '''
            the length of dataframe is based on:
            
                * the num of `ids` if there is more than one id
                * the num of `sites` if there is just one id with specified `sites`
            '''
            if isinstance(ids, str):
                if sites is not None and not isinstance(sites, str):
                    index_len = len(sites)
                else:
                    index_len = 1
            else:
                index_len = len(ids)

            self.dfrm = DataFrame(dict(zip(
                (col for col in (id_col, site_col, gene_col) if col is not None),
                (value for value in (ids, sites, genes) if value is not None))),
                index=list(range(index_len)))

        self.index = dfrm.index
        self.id_col = id_col
        self.id_type = id_type
        self.site_col = site_col
        self.gene_col = gene_col
        self.loggingPath = loggingPath
        if isinstance(usecols, str):
            PARAMS['columns'] = usecols
            usecols = usecols.split(',')
        elif isinstance(usecols, (Iterable, Iterator)):
            PARAMS['columns'] = ','.join(usecols)
        else:
            raise ValueError('Invalid usecols')
        self.usecols = usecols
        PARAMS['from'] = id_type
        if isinstance(loggingPath, (str, Path)):
            self.set_logging_fileHandler(loggingPath)

    @property
    def sites(self) -> Generator:
        if self.site_col is not None:
            for name, group in self.dfrm.groupby(by=self.id_col, sort=False):
                yield name, group[self.site_col]
        else:
            yield None

    @staticmethod
    def split_df(dfrm, colName, sep):
        '''Split DataFrame'''
        df = dfrm.copy()
        return df.drop([colName], axis=1).join(df[colName].str.split(sep, expand=True).stack().reset_index(level=1, drop=True).rename(colName))

    def yieldTasks(self, lyst: Iterable, chunksize: int = 100, sep: str = ',') -> Generator:
        fileName = self.outputPath.stem
        for i in range(0, len(lyst), chunksize):
            cur_fileName = f'{fileName}+{i}'
            cur_params = deepcopy(PARAMS)
            cur_params['query'] = sep.join(lyst[i:i+chunksize]) # self.outputPath.suffix
            yield ('get', {'url': f'{BASE_URL}/uploadlists/', 'params': cur_params}, str(Path(self.outputPath.parent, cur_fileName+'.tsv')))

    def retrieve(self, outputPath: Union[str, Path], 
                       finishedPath: Optional[str] = None, 
                       sep: str = '\t', 
                       chunksize: int = 100, 
                       concur_req: int = 20, 
                       rate: float = 1.5,
                       ret_res: bool = True,
                       semaphore = None):
        finish_id = list()
        self.outputPath = Path(outputPath)
        self.result_cols = [COLUMNS_DICT.get(
            i, i) for i in self.usecols] + RESULT_NEW_COLUMN
        if finishedPath is not None:
            try:
                target_col = RESULT_NEW_COLUMN[0]
                finish: Series = read_csv(
                    finishedPath,
                    sep=sep,
                    usecols=[target_col],
                    names=self.result_cols,
                    skiprows=1,
                    header=None)[target_col]
            except Exception as e:
                col_to_add = RESULT_NEW_COLUMN[1]
                self.logger.warning(
                    f"{e}\nSomething wrong with finished raw file, probably without '{col_to_add}' column.")
                finish_df = read_csv(
                    finishedPath, sep=sep, names=self.result_cols[:-1], skiprows=1, header=None)
                finish_df[col_to_add] = nan
                finish_df.to_csv(finishedPath, sep=sep, index=False)
                finish: Series = finish_df[target_col]

            for query_id in finish:
                if ',' in query_id:
                    finish_id.extend(query_id.split(','))
                else:
                    finish_id.append(query_id)

        query_id: Series = self.dfrm[self.id_col]
        if finish_id:
            rest_id = list(set(query_id) - set(finish_id))
        else:
            rest_id = query_id.unique()

        self.logger.info(
            f"Have finished {len(finish_id)} ids, {len(rest_id)} ids left.")
        res = UnsyncFetch.multi_tasks(
            tasks=self.yieldTasks(rest_id, chunksize),
            to_do_func=self.process,
            concur_req=concur_req, 
            rate=rate, 
            ret_res=ret_res,
            semaphore=semaphore)
        return res

    def getCanonicalInfo(self, dfrm: DataFrame):
        '''
        Will Change the dfrm

        * Add new column (canonical_isoform)
        * Change the content of column (UniProt)
        '''
        # Get info from Alt Product file
        if self.altProPath is None:
            dfrm['canonical_isoform'] = nan
            return dfrm
        else:
            usecols = ["IsoId", "Sequence", "Entry", "UniProt"]
            altPro_df = read_csv(self.altProPath, sep="\t", usecols=usecols)
            altPro_df = altPro_df[altPro_df["Sequence"]
                                  == "Displayed"].reset_index(drop=True)
            altPro_df.rename(
                columns={"IsoId": "canonical_isoform"}, inplace=True)
            # Modify dfrm
            dfrm = merge(
                dfrm, altPro_df[["canonical_isoform", "Entry"]], how="left")
            return dfrm

    def getGeneStatus(self, handled_df: DataFrame, colName: str = 'GENE_status'):
        '''
        Will Change the dfrm, add Gene Status

        * Add new column (GENE) # if id_col != gene_col
        * Add new column (GENE_status)

        **About GENE_status**

        * ``False`` : First element of Gene names is not correspond with refSeq's GENE (e.g)
        * others(corresponding GENE)

        '''
        self.gene_status_col = colName
        if self.id_type != 'GENENAME':
            if self.gene_col is None:
                handled_df[colName] = True
                return None
            gene_map = self.dfrm[[self.id_col,
                                  self.gene_col]].drop_duplicates()
            gene_map = gene_map.groupby(self.id_col)[self.gene_col].apply(
                lambda x: array(x) if len(x) > 1 else list(x)[0])
            handled_df['GENE'] = handled_df.apply(
                lambda z: gene_map[z['yourlist']], axis=1)
            handled_df[colName] = handled_df.apply(lambda x: x['GENE'] == x['Gene names'].split(
                ' ')[0] if not isinstance(x['Gene names'], float) else False, axis=1)
            handled_df['GENE'] = handled_df['GENE'].apply(
                lambda x: ','.join(x) if not isinstance(x, str) else x)
        else:
            handled_df[colName] = handled_df.apply(lambda x: x['yourlist'] == x['Gene names'].split(
                ' ')[0] if not isinstance(x['Gene names'], float) else False, axis=1)

    def label_mapping_status(self, dfrm: DataFrame, colName: str = 'Mapping_status'):
        self.mapping_status_col = colName
        gene_status_col = self.gene_status_col
        dfrm[colName] = 'No'
        dfrm[gene_status_col] = dfrm[gene_status_col].apply(
            lambda x: x.any() if isinstance(x, Iterable) else x)

        if self.id_col == 'GENENAME':
            pass_df = dfrm[
                (dfrm[gene_status_col] == True) &
                (dfrm['Status'] == 'reviewed') &
                (dfrm['unp_map_tage'] != 'Untrusted & No Isoform')]
        else:
            pass_df = dfrm[
                (dfrm['Status'] == 'reviewed') &
                (dfrm['unp_map_tage'] != 'Untrusted & No Isoform')]
        pass_index = pass_df.index
        dfrm.loc[pass_index, colName] = 'Yes'

        # Deal with 'one to many' situation
        multipleCounter = Counter(dfrm.loc[pass_index, 'yourlist'])
        err_li = [i for i, j in multipleCounter.items() if j > 1]
        err_index = pass_df[pass_df['yourlist'].isin(err_li)].index
        dfrm.loc[err_index, colName] = 'Error'

    @unsync
    async def process(self, path: Union[str, Path, Unfuture], sep: str = '\t'):
        self.logger.debug("Start to handle id mapping result")
        if not isinstance(path, (Path, str)):
            path = await path  # .result()
        if not Path(path).stat().st_size:
            return None
        self.altSeqPath, self.altProPath = ExtractIsoAlt.main(path=path)
        try:
            df = read_csv(
                path, sep='\t', names=self.result_cols, skiprows=1, header=None)
        except ValueError:
            df = read_csv(
                path, sep='\t', names=self.result_cols[:-1], skiprows=1, header=None)

        # Add New Column: canonical_isoform
        df = self.getCanonicalInfo(df)
        # Add New Column: unp_map_tage
        df['unp_map_tage'] = nan
        # Classification
        df_with_no_isomap = df[df['isomap'].isnull()]  # Class A
        df_with_isomap = df[df['isomap'].notnull()]  # Class B
        # ----------------------------------------------------------------------
        # In Class A
        # ----------------------------------------------------------------------
        if len(df_with_no_isomap) > 0:
            df_wni_split = self.split_df(df_with_no_isomap, 'yourlist', ',')
            df_wni_split.drop(columns=['isomap'], inplace=True)
            # [yourlist <-> UniProt]
            df_wni_split['UniProt'] = df_wni_split['Entry']
            df_wni_split['unp_map_tage'] = 'Trusted & No Isoform'
            # Find out special cases 1
            df_wni_split_warn = df_wni_split[df_wni_split['Alternative products (isoforms)'].notnull(
            )].index
            df_wni_split.loc[df_wni_split_warn,
                             'unp_map_tage'] = 'Untrusted & No Isoform'
            # 'Entry', 'Gene names', 'Status', 'Alternative products (isoforms)', 'Organism', 'yourlist', 'UniProt'
        # ----------------------------------------------------------------------
        # In Class B
        # ----------------------------------------------------------------------
        if len(df_with_isomap) > 0:
            wi_yourlist_count = df_with_isomap.apply(
                lambda x: x['yourlist'].count(','), axis=1)
            wi_isomap_count = df_with_isomap.apply(
                lambda x: x['isomap'].count(','), axis=1)
            # In subClass 1
            df_wi_eq = df_with_isomap.loc[wi_yourlist_count[wi_yourlist_count ==
                                                            wi_isomap_count].index]
            if len(df_wi_eq) > 0:
                df_wi_eq_split = self.split_df(
                    df_wi_eq.drop(columns=['yourlist']), 'isomap', ',')
                df_wi_eq_split[['yourlist', 'UniProt']] = df_wi_eq_split['isomap'].str.split(
                    ' -> ', expand=True)
                # [yourlist <-> UniProt]
                df_wi_eq_split.drop(columns=['isomap'], inplace=True)
                df_wi_eq_split['unp_map_tage'] = 'Trusted & Isoform'
                # # 'Entry', 'Gene names', 'Status', 'Alternative products (isoforms)', 'Organism', 'yourlist', 'UniProt'

            # In subClass 2
            df_wi_ne = df_with_isomap.loc[wi_yourlist_count[wi_yourlist_count !=
                                                            wi_isomap_count].index]
            if len(df_wi_ne) > 0:
                df_wi_ne_split = self.split_df(df_wi_ne, 'isomap', ',')
                df_wi_ne_split.rename(
                    columns={'yourlist': 'checkinglist'}, inplace=True)
                df_wi_ne_split[['yourlist', 'UniProt']] = df_wi_ne_split['isomap'].str.split(
                    ' -> ', expand=True)
                df_wi_ne_split.drop(columns=['isomap'], inplace=True)
                df_wi_ne_split['unp_map_tage'] = 'Trusted & Isoform & Contain Warnings'
                # 'Entry', 'Gene names', 'Status', 'Alternative products (isoforms)', 'Organism', 'yourlist', 'UniProt', 'checkinglist'
                # Find out special cases 2
                usecols = Index(set(df_wi_ne_split.columns) -
                                   {'yourlist', 'UniProt'})
                df_wi_ne_warn = self.split_df(
                    df_wi_ne_split[usecols].drop_duplicates(), 'checkinglist', ',')
                df_wi_ne_warn = df_wi_ne_warn[~df_wi_ne_warn['checkinglist'].isin(
                    df_wi_ne_split['yourlist'])].rename(columns={'checkinglist': 'yourlist'})
                df_wi_ne_warn['UniProt'] = df_wi_ne_warn['Entry']
                # sequence conflict
                df_wi_ne_warn['unp_map_tage'] = 'Untrusted & No Isoform'
                df_wi_ne_split.drop(columns=['checkinglist'], inplace=True)

        # Concat Dfrm
        variables = ["df_wni_split", "df_wi_eq_split",
                     "df_wi_ne_split", "df_wi_ne_warn"]
        lvs = locals()
        varLyst = [lvs[variable] for variable in variables if variable in lvs]
        final_df = concat(varLyst, sort=False).reset_index(drop=True)
        cano_index = final_df[final_df["canonical_isoform"].notnull()].index
        if len(cano_index) > 0:
            final_df.loc[cano_index, "UniProt"] = final_df.loc[cano_index, ].apply(
                lambda x: x["Entry"] if x["UniProt"] in x["canonical_isoform"] else x["UniProt"], axis=1)

        # Add Gene Status
        self.getGeneStatus(final_df)
        # Label Mapping Status
        self.label_mapping_status(final_df)

        pathOb = Path(path)
        edPath = str(Path(pathOb.parent, f'{pathOb.stem}_ed.tsv'))  # {pathOb.suffix}
        final_df.to_csv(edPath, sep=sep, index=False)
        self.logger.debug(f"Handled id mapping result saved in {edPath}")
        return edPath
"""

class UniProtAPI(Abclog):
    '''
    Implement UniProt Retrieve/ID Mapping API

    * focus on tabular format
    * <https://www.uniprot.org/help/uploadlists>
    * <https://www.uniprot.org/help/api_idmapping>
    '''
    params = {
        'columns': 'id,feature(ALTERNATIVE%20SEQUENCE)',
        'query': None,
        'from': 'ACC+ID',
        'to': 'ACC',
        'format': 'tab'}

    with_name_suffix = True

    @classmethod
    def task_unit(cls, chunk, i, folder, name, sep):
        cur_params = deepcopy(cls.params)
        cur_params['query'] = sep.join(chunk)
        return ('get', {'url': f'{BASE_URL}/uploadlists/', 'params': cur_params}, folder/f'{name}+{i}.tsv')

    @classmethod
    def yieldTasks(cls, lyst: Iterable, chunksize: int, folder, name: str, sep: str = ',') -> Generator:
        name_with_suffix = f'{name}+{uuid4().hex}' if cls.with_name_suffix else name
        for i in range(0, len(lyst), chunksize):
            yield cls.task_unit(lyst[i:i+chunksize], i, folder, name_with_suffix, sep)
    
    @classmethod
    @unsync
    async def set_web_semaphore(cls, web_semaphore_value: int):
        cls.web_semaphore = await init_semaphore(web_semaphore_value)
    
    @classmethod
    def set_folder(cls, folder: Union[Path, str]):
        cls.folder = init_folder_from_suffix(folder, 'UniProt/uploadlists/')
    
    @classmethod
    def retrieve(cls, lyst: Iterable, name: str, sep: str = ',', chunksize: int = 100, rate: float = 1.5, semaphore=None, **kwargs):
        return [UnsyncFetch.single_task(
            task,
            semaphore=cls.web_semaphore if semaphore is None else semaphore,
            rate=rate) for task in cls.yieldTasks(lyst, chunksize, cls.folder, name, sep)]


class UniProtFASTA(Abclog):
    '''
    Download UniProt Fasta Sequences

    >>> UniProtFASTA.retrieve(
        ('Q6NZ36', 'P12755'),
        init_folder_from_suffix(yourfolder, 'UniProt/fasta/'))
    '''

    params = {'include': 'no'}
    obj = {}

    @classmethod
    @unsync
    async def set_web_semaphore(cls, web_semaphore_value:int):
        cls.web_semaphore = await init_semaphore(web_semaphore_value)
    
    @classmethod
    def set_folder(cls, folder: Union[Path, str]):
        cls.folder = init_folder_from_suffix(folder, 'UniProt/fasta/')

    """
    @classmethod
    @unsync
    async def process(cls, path: Union[str, Path, Unfuture]):
        if not isinstance(path, (Path, str)):
            path = await path  # .result()
        path = Path(path)
        if not path.stat().st_size:
            return None
        if not cls.obj:
            folder = path.parent
            kwargs = {'ret_res': False}
        else:
            folder = cls.obj['fasta_folder']
            kwargs = {'concur_req': cls.obj['unp_concurreq'], 'rate': cls.obj['unp_concurrate'],
                      'ret_res': False, 'semaphore': cls.obj['semaphore']}
        unps = read_csv(path, sep='\t', usecols=['UniProt']).UniProt.drop_duplicates()
        for fob in cls.retrieve(unps, folder, **kwargs):
            await fob
        return path
        """

    @classmethod
    def task_unit(cls, unp:str, folder: Union[str, Path]):
        cur_fileName = f'{unp}.fasta'
        cur_filePath = str(Path(folder, cur_fileName))
        return ('get', {'url': f'{BASE_URL}/uniprot/{cur_fileName}', 'params': cls.params}, cur_filePath)

    @classmethod
    def yieldTasks(cls, lyst: Iterable, folder: Union[str, Path]) -> Generator:
        for unp in lyst:
            return cls.task_unit(unp, folder)

    @classmethod
    def retrieve(cls, lyst: Iterable, folder: Union[str, Path], concur_req: int = 20, rate: float = 1.5, ret_res: bool = True, semaphore=None):
        return UnsyncFetch.multi_tasks(
            cls.yieldTasks(lyst, folder), 
            concur_req=concur_req, 
            rate=rate, 
            ret_res=ret_res,
            semaphore=cls.web_semaphore if semaphore is None else semaphore)
    
    @classmethod
    def single_retrieve(cls, identifier: str, folder: Optional[Union[str, Path]]=None, semaphore=None, rate: float = 1.5):
        return UnsyncFetch.single_task(
            task=cls.task_unit(identifier, cls.folder if folder is None else folder),
            semaphore=cls.web_semaphore if semaphore is None else semaphore,
            rate=rate)



# @Created Date: 2020-09-26 02:51:49 pm
# @Filename: api.py
# @Email:  1730416009@stu.suda.edu.cn
# @Author: ZeFeng Zhu
# @Last Modified: 2020-09-26 02:51:59 pm
# @Copyright (c) 2020 MinghuiGroup, Soochow University
from typing import Dict, Tuple, Iterable, Generator, Union
from pathlib import Path
from pdb_profiling.log import Abclog
from pdb_profiling.fetcher.webfetch import UnsyncFetch

BASE_URL = 'https://eutils.ncbi.nlm.nih.gov/entrez/eutils/'


class EutilsAPI(Abclog):
    '''
    Implement The Entrez Programming Utilities (E-utilities) API

    * Entrez Programming Utilities Help: <https://www.ncbi.nlm.nih.gov/books/NBK25501/>
    * DEMO URL 1: <https://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi?db=protein&id=NP_001291289&rettype=fasta>
    * DEMO URL 2: <https://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi?db=sequences&id=NM_001304360&rettype=fasta>
    '''
    headers = {"Content-Type": "text/plain"}
    api_set = frozenset(('efetch.fcgi', 'einfo.fcgi', 'esearch.fcgi',
                         'epost.fcgi', 'esummary.fcgi', 'egquery.fcgi'))

    @classmethod
    def dumpsParams(cls, params: Dict) -> str:
        return '&'.join(f'{key}={value}' for key, value in params.items())

    @classmethod
    def task_unit(cls, suffix: str, params: Dict, folder: Path) -> Tuple:
        args = dict(
            url=f'{BASE_URL}{suffix}',
            headers=cls.headers,
            params=params)
        return 'get', args, folder/f'{cls.dumpsParams(params)}.{params.get("retmode", params.get("rettype", "xml"))}'

    @classmethod
    def yieldTasks(cls, suffix: str, params_collection: Iterable[Dict], folder: Path) -> Generator:
        for params in params_collection:
            yield cls.task_unit(suffix, params, folder)

    @classmethod
    def retrieve(cls, suffix: str, params_collection: Iterable[Dict], folder: Union[Path, str], concur_req: int = 20, rate: float = 1.5, ret_res: bool = True, **kwargs):
        assert suffix in cls.api_set, f"Invalid suffix! Valid set is \n{cls.api_set}"
        folder = Path(folder)
        res = UnsyncFetch.multi_tasks(
            cls.yieldTasks(suffix, params_collection, folder),
            concur_req=concur_req,
            rate=rate,
            ret_res=ret_res,
            semaphore=kwargs.get('semaphore', None))
        return res

    @classmethod
    def single_retrieve(cls, suffix: str, params: Dict, folder: Union[Path, str], semaphore, rate: float = 1.5):
        assert suffix in cls.api_set, f"Invalid suffix! Valid set is \n{cls.api_set}"
        folder = Path(folder)
        return UnsyncFetch.single_task(
            task=cls.task_unit(suffix, params, folder),
            semaphore=semaphore,
            rate=rate)



# @Created Date: 2020-10-21 09:09:05 pm
# @Filename: viewer.py
# @Email:  1730416009@stu.suda.edu.cn
# @Author: ZeFeng Zhu
# @Last Modified: 2020-10-21 09:09:10 pm
# @Copyright (c) 2020 MinghuiGroup, Soochow University
from pandas import isna, DataFrame
from pdb_profiling.utils import expand_interval
from pdb_profiling.processors.pdbe.record import PDB


class NGL(object):

    @staticmethod
    def get_related_auth_res(res_df, struct_asym_id, res_num_set):
        if len(res_num_set) > 0:
            subset = res_df[res_df.struct_asym_id.eq(struct_asym_id) & res_df.residue_number.isin(res_num_set)]
            return ('('+subset.author_residue_number.astype(str)+'^'+subset.author_insertion_code+subset.multiple_conformers.apply(lambda x: ' and %' if isna(x) else ' and %A')+')')
        else:
            subset = res_df[res_df.struct_asym_id.eq(struct_asym_id)]
            assert len(subset) == 1
            return ('('+subset.author_residue_number.astype(str)+'^'+subset.author_insertion_code+subset.multiple_conformers.apply(lambda x: ' and %' if isna(x) else ' and %A')+')').iloc[0]

    @staticmethod
    def get_type(x):
        if x in (
            'polypeptide(L)',
            'polypeptide(D)'):
            return 'protein'
        elif x in (
            'polydeoxyribonucleotide',
            'polyribonucleotide',
            'polydeoxyribonucleotide/polyribonucleotide hybrid'):
            return 'nucleic'
        else:
            return 'ligand'

    @classmethod
    def interface_sele_str_unit(cls, res_df, record, suffix):
        mol_type = cls.get_type(record[f'molecule_type{suffix}'])
        if mol_type != 'ligand':
            res = cls.get_related_auth_res(res_df, record[f'struct_asym_id{suffix}'], frozenset(expand_interval(record[f'interface_range{suffix}'])))
            sres = cls.get_related_auth_res(res_df, record[f'struct_asym_id{suffix}'], frozenset(expand_interval(record[f'surface_range{suffix}'])))
            res = ' or '.join(res)
            sres = ' or '.join(sres)
            return mol_type, f' and ({res})', f' and ({sres})'
        else:
            res = cls.get_related_auth_res(res_df, record[f'struct_asym_id{suffix}'], frozenset())
            sres = cls.get_related_auth_res(res_df, record[f'struct_asym_id{suffix}'], frozenset())
            return mol_type, f' and {res}', f' and {sres}'

    @classmethod
    def get_interface_sele_str(cls, record):
        res_df = PDB(record['pdb_id']).fetch_from_web_api('api/pdb/entry/residue_listing/', PDB.to_dataframe).result()
        type_1, i_str_1, s_str_1 = cls.interface_sele_str_unit(res_df, record, '_1')
        type_2, i_str_2, s_str_2 = cls.interface_sele_str_unit(res_df, record, '_2')
        chain_id_1 = record['chain_id_1']
        chain_id_2 = record['chain_id_2']
        return ((f'{type_1} and :{chain_id_1}'+i_str_1, f'{type_2} and :{chain_id_2}'+i_str_2),
                (f'{type_1} and :{chain_id_1}'+s_str_1, f'{type_2} and :{chain_id_2}'+s_str_2))

    @classmethod
    def get_interface_view(cls, view, record, **kwargs):
        (i1, i2), (s1, s2) = cls.get_interface_sele_str(record)
        view.add_spacefill(selection=s1, opacity=kwargs.get('surface_opacity_1', 0.05), color=kwargs.get('surface_color_1', 'white'))
        view.add_spacefill(selection=s2, opacity=kwargs.get('surface_opacity_2', 0.05), color=kwargs.get('surface_color_2', 'white'))
        view.add_spacefill(selection=i1, opacity=kwargs.get('interface_opacity_1', 0.5), color=kwargs.get('interface_color_1', 'green'))
        view.add_spacefill(selection=i2, opacity=kwargs.get('interface_opacity_2', 0.5), color=kwargs.get('interface_color_2', 'red'))
        view.background = kwargs.get('background', '#F3F3F3')
        view._set_size(*kwargs.get('size', ('50%', '50%')))
        return view



# @Created Date: 2020-10-17 08:02:13 am
# @Filename: __init__.py
# @Email:  1730416009@stu.suda.edu.cn
# @Author: ZeFeng Zhu
# @Last Modified: 2020-10-17 08:02:20 am
# @Copyright (c) 2020 MinghuiGroup, Soochow University
import orm
from unsync import unsync
from pdb_profiling.processors.database import SqliteDB


class I3DDB(SqliteDB):

    def init_table_model(self):

        class InteractionMeta(orm.Model):
            __tablename__ = 'InteractionMeta'
            __metadata__ = self.metadata
            __database__ = self.database
            pdb_id = orm.String(max_length=4, primary_key=True)
            Entry_1 = orm.String(max_length=100, primary_key=True)
            Entry_2 = orm.String(max_length=100, primary_key=True)
            assembly_id = orm.Integer(primary_key=True)
            chain_id_1 = orm.String(max_length=4, primary_key=True)
            model_id_1 = orm.Integer(primary_key=True)
            chain_id_2 = orm.String(max_length=4, primary_key=True)
            model_id_2 = orm.Integer(primary_key=True)
            organism = orm.String(max_length=100)
            interaction_type = orm.String(max_length=2)

        self.InteractionMeta = InteractionMeta



# @Created Date: 2020-11-23 10:29:17 am
# @Filename: command.py
# @Email:  1730416009@stu.suda.edu.cn
# @Author: ZeFeng Zhu
# @Last Modified: 2020-11-23 10:29:36 am
# @Copyright (c) 2020 MinghuiGroup, Soochow University
from pdb_profiling import default_config
from pdb_profiling.commands import CustomDB
from pdb_profiling.processors import Identifiers, Identifier, SIFTSs
from pdb_profiling.utils import unsync_run
import click
from pandas import read_csv
from importlib import util as imp_util
from pathlib import Path
from math import ceil
from tqdm.auto import tqdm
from time import sleep
from random import uniform


def colorClick(name: str, template: str = "Initializing %s", fg: str = "green"):
    return click.style(template % name, fg=fg)


@click.group(chain=True, invoke_without_command=False)
@click.option("--folder", default="./", help="The output folder.", type=click.Path())
@click.option('--dropall/--no-dropall', help="whether to use existing custom DB", default=False, is_flag=True)
@click.pass_context
def Interface(ctx, folder, dropall):
    folder = Path(folder)
    click.echo(colorClick(f"Folder: {folder}"))
    ctx.ensure_object(dict)
    ctx.obj['folder'] = folder
    default_config(folder)
    ctx.obj['custom_db'] = CustomDB(
        "sqlite:///%s" % (folder/'local_db'/'custom.db'), dropall)


@Interface.command("insert-mutation")
@click.option("--input", help="the file that contains sites info", type=click.Path())
@click.option("--sep", default="\t", help="the seperator of input file", type=str)
@click.option("--usecols", default='from_id,Ref,Pos,Alt', help="The comma-sep columns of site info", type=str)
@click.option('--readchunk', type=int, help="the chunksize parameter of pandas.read_csv", default=100000)
@click.option('--nrows', type=int, help="the nrows parameter of pandas.read_csv", default=None)
@click.option('--skiprows', type=int, help="the skiprows parameter of pandas.read_csv", default=None)
@click.option('--functionfile', help="the py file that contains custom function", default=None, type=click.Path())
@click.option('--functionname', default='do_something', type=str)
@click.pass_context
def insert_sites(ctx, input, sep, usecols, readchunk, nrows, skiprows, functionfile, functionname):
    def do_nothing(dfrm):
        return dfrm.to_dict('records')

    click.echo(colorClick("DB Mutation Insertion"))
    usecols = usecols.split(',')
    if functionfile is not None:
        spec = imp_util.spec_from_file_location("CustomFunc", functionfile)
        CustomFunc = imp_util.module_from_spec(spec)
        spec.loader.exec_module(CustomFunc)
        deal = getattr(CustomFunc, functionname)
        click.echo(f"Success: load func: {functionname} from {functionfile}")
    else:
        deal = do_nothing
    df = read_csv(input, sep=sep, usecols=usecols, chunksize=readchunk,
                  nrows=nrows, skiprows=skiprows)
    sqlite_api = ctx.obj['custom_db']
    start = 0
    for index, dfrm in enumerate(df):
        end = readchunk*(index+1)
        click.echo(f"Try to insert: {start}-{end}")
        start = end+1
        sqlite_api.sync_insert(sqlite_api.Mutation, deal(dfrm))


@Interface.command("id-mapping")
@click.option('--chunksize', type=int, help="the chunksize parameter", default=500)
@click.pass_context
def id_mapping(ctx, chunksize):
    sqlite_api = ctx.obj['custom_db']
    total = unsync_run(sqlite_api.database.fetch_one(
        query="SELECT COUNT(DISTINCT ftId) FROM Mutation"))[0]
    cols = ('ftId', 'Entry', 'isoform', 'is_canonical')
    click.echo(f"Total {total} to query")
    for i in range(ceil(total/chunksize)):
        res = unsync_run(sqlite_api.database.fetch_all(query=f"""
            SELECT DISTINCT ftId FROM Mutation
            WHERE ftId NOT IN (SELECT DISTINCT ftId FROM IDMapping)
            LIMIT {chunksize} OFFSET {chunksize*i}
            """))
        res = Identifiers(i[0] for i in res).fetch(
            'map2unp').run(tqdm).result()
        values = [dict(zip(cols, i)) for i in res if i is not None]
        if values:
            sqlite_api.sync_insert(sqlite_api.IDMapping, values)
        sleep(uniform(1, 10))


@Interface.command("sifts-mapping")
@click.option('--func', type=str, default='pipe_select_mo')
@click.option('--chunksize', type=int, help="the chunksize parameter", default=200)
@click.option('--output', type=str)
@click.pass_context
def sifts_mapping(ctx, func, chunksize, output):
    def get_unp_id(args):
        if args is None:
            return
        Entry, isoform, is_canonical = args
        return Entry if is_canonical and isoform != 'NaN' else isoform

    sqlite_api = ctx.obj['custom_db']
    total = unsync_run(sqlite_api.database.fetch_one(
        query="SELECT COUNT(DISTINCT isoform) FROM IDMapping"))[0]
    click.echo(f"Total {total} to query")
    for i in range(ceil(total/chunksize)):
        res = unsync_run(sqlite_api.database.fetch_all(
            query=f"SELECT DISTINCT Entry,isoform,is_canonical FROM IDMapping LIMIT {chunksize} OFFSET {chunksize*i}"))
        res = SIFTSs(i for i in map(get_unp_id, res)
                     if i is not None).fetch(func).run(tqdm).result()
        output_path = ctx.obj['folder']/output
        for dfrm in res:
            dfrm.to_csv(output_path, sep='\t', index=False,
                        header=not output_path.exists())
        sleep(uniform(1, 10))


'''
@Interface.command("residue-mapping")
@click.option('--chunksize', type=int, help="the chunksize parameter", default=200)
@click.option('--output', type=str)
@click.pass_context
def sifts_mapping(ctx, func, chunksize, output):
'''


if __name__ == '__main__':
    Interface(obj={})



# @Created Date: 2020-08-12 11:28:01 pm
# @Filename: __init__.py
# @Email:  1730416009@stu.suda.edu.cn
# @Author: ZeFeng Zhu
# @Last Modified: 2020-10-10 10:37:46 am
# @Copyright (c) 2020 MinghuiGroup, Soochow University
import orm
from unsync import unsync
from re import compile as re_compile
from pdb_profiling.processors.database import SqliteDB

common_pat = r'^(?=.*[A-Za-z])(?=.*\d)[A-Za-z\d]'


pats = dict(pdb_id=re_compile(common_pat+r'{4}$'),
            pdb_entity_id=re_compile(common_pat+r'{4}_[0-9]+$'),
            UniProt=re_compile(common_pat+r'{6,}[\-]*[0-9]*$'),
            pdb_complex_id=re_compile(r'PDB-CPX-[0-9]+'))


def default_id_tag(identifier: str, default: str = '', raise_error: bool = False):
    try:
        for pat_name, pat in pats.items():
            if bool(pat.fullmatch(identifier)):
                return pat_name
    except Exception:
        raise ValueError(f"Invalid Identifier: {identifier} !")
    if raise_error:
        raise ValueError(f'Unexpected Identifiers: {identifier}')
    else:
        return default


class PDBeDB(SqliteDB):

    def init_table_model(self):
        class ResidueMapping(orm.Model):
            __tablename__ = 'ResidueMapping'
            __metadata__ = self.metadata
            __database__ = self.database
            UniProt = orm.String(max_length=50, primary_key=True)
            author_insertion_code = orm.String(max_length=50, allow_null=True, allow_blank=True, default='')
            author_residue_number = orm.Integer()
            chain_id = orm.String(max_length=10)
            entity_id = orm.Integer(primary_key=True)
            identifier = orm.Text(allow_null=True, allow_blank=True, default='')
            name = orm.Text(allow_null=True, allow_blank=True, default='')
            observed = orm.String(max_length=1)
            pdb_id = orm.String(max_length=4, primary_key=True)
            pdb_one_letter_code = orm.String(max_length=1, primary_key=True)
            residue_number = orm.Integer(primary_key=True)
            struct_asym_id = orm.String(max_length=10, primary_key=True)
            unp_one_letter_code = orm.String(max_length=1, primary_key=True)
            unp_residue_number = orm.Integer(primary_key=True)

        class StatsProteinEntitySeq(orm.Model):
            __tablename__ = 'StatsProteinEntitySeq'
            __metadata__ = self.metadata
            __database__ = self.database
            pdb_id = orm.String(max_length=4, primary_key=True)
            molecule_type = orm.Text()
            entity_id = orm.Integer(primary_key=True)
            ca_p_only = orm.Boolean()
            SEQRES_COUNT = orm.Integer()
            STD_INDEX = orm.JSON()
            STD_COUNT = orm.Integer()
            NON_INDEX = orm.JSON()
            NON_COUNT = orm.Integer()
            UNK_INDEX = orm.JSON()
            UNK_COUNT = orm.Integer()
            ARTIFACT_INDEX = orm.JSON()
        
        class StatsNucleotideEntitySeq(orm.Model):
            __tablename__ = 'StatsNucleotideEntitySeq'
            __metadata__ = self.metadata
            __database__ = self.database
            pdb_id = orm.String(max_length=4, primary_key=True)
            molecule_type = orm.Text()
            entity_id = orm.Integer(primary_key=True)
            ca_p_only = orm.Boolean()
            SEQRES_COUNT = orm.Integer()
            dNTP_INDEX = orm.JSON()
            dNTP_COUNT = orm.Integer()
            NTP_INDEX = orm.JSON()
            NTP_COUNT = orm.Integer()
            NON_INDEX = orm.JSON()
            NON_COUNT = orm.Integer()
            UNK_INDEX = orm.JSON()
            UNK_COUNT = orm.Integer()

        class StatsChainSeq(orm.Model):
            __tablename__ = 'StatsChainSeq'
            __metadata__ = self.metadata
            __database__ = self.database
            pdb_id = orm.String(max_length=4, primary_key=True)
            entity_id = orm.Integer(primary_key=True)
            chain_id = orm.String(max_length=10)
            struct_asym_id = orm.String(max_length=10, primary_key=True)
            OBS_INDEX = orm.JSON()
            OBS_COUNT = orm.Integer()
            OBS_RATIO_ARRAY = orm.Text()
            BINDING_LIGAND_INDEX = orm.JSON()
            BINDING_LIGAND_COUNT = orm.Integer()

        self.ResidueMapping = ResidueMapping
        self.StatsProteinEntitySeq = StatsProteinEntitySeq
        self.StatsNucleotideEntitySeq = StatsNucleotideEntitySeq
        self.StatsChainSeq = StatsChainSeq
